# Copyright 2022 Cognite AS
import re
import warnings

from datetime import datetime
from typing import Callable, List, Literal, Optional, Union

import numpy as np
import pandas as pd
import scipy.integrate

from indsl.type_check import check_types

from ..exceptions import UserRuntimeError, UserValueError


_unit_in_ms_without_week = {"s": 1000, "m": 60000, "h": 3600000, "d": 86400000}
_unit_in_ms = {**_unit_in_ms_without_week, "w": 604800000}


def datetime_to_ms(dt):
    """Converts datetime to timestamp in milliseconds."""
    epoch = datetime.utcfromtimestamp(0)
    return int((dt - epoch).total_seconds() * 1000.0)


def time_string_to_ms(pattern, string, unit_in_ms):
    """Converts time from string to milliseconds representation."""
    pattern = pattern.format("|".join(unit_in_ms))
    res = re.fullmatch(pattern, string)
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        return magnitude * unit_in_ms[unit]
    return None


def granularity_to_ms(granularity: str) -> int:
    """Converts granularity from string to milliseconds representation."""
    ms = time_string_to_ms(r"(\d+)({})", granularity, _unit_in_ms_without_week)
    if ms is None:
        raise UserValueError(
            "Invalid granularity format: `{}`. Must be on format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".format(
                granularity
            )
        )
    return ms


def time_ago_to_ms(time_ago_string: str) -> int:
    """Returns millisecond representation of time-ago string."""
    if time_ago_string == "now":
        return 0
    ms = time_string_to_ms(r"(\d+)({})-ago", time_ago_string, _unit_in_ms)
    if ms is None:
        raise UserValueError(
            "Invalid time-ago format: `{}`. Must be on format <integer>(s|m|h|d|w)-ago or 'now'. E.g. '3d-ago' or '1w-ago'.".format(
                time_ago_string
            )
        )
    return ms


def get_fixed_freq(
    start_time: Union[datetime, str, int, float],
    end_time: Union[datetime, str, int, float],
    granularity: Union[str, pd.Timedelta],
):
    """Creates a time series index with a fixed frequency.

    Args:
        start_time (Union(pandas.Datetime,str, int, float)): Start time value to be converted to Timestamp.
        end_time (Union(pandas.Datetime,str, int, float)): End time value to be converted to Timestamp.
        granularity (str): Offset which Timestamp will have. Must be in format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".

    Returns:
        pd.DatetimeIndex: DatetimeIndex with fixed frequency.
    """
    if isinstance(granularity, str):
        return pd.date_range(start_time, end_time, freq=pd.Timedelta(milliseconds=granularity_to_ms(granularity)))
    else:
        return pd.date_range(start_time, end_time, freq=granularity)


def above_below(series: pd.Series, upper_limit: float, lower_limit: float):
    """Above/below.

    Counts the number of points above/below upper_limit/lower_limit.

    Args:
        series: Data with a pandas.DatetimeIndex.
        upper_limit: Value for upper limit.
        lower_limit: Value for lower limit.

    Returns:
        dict: The numbers above/below upper_limit/lower_limit.
    """
    result = {}
    if upper_limit is not None:
        result["above_upper"] = (series > upper_limit).sum()

    if lower_limit is not None:
        result["below_lower"] = (series < lower_limit).sum()

    return result


def num_vals_in_boxes(series: pd.Series, num_boxes: int):
    """Counts the number of points in each of the boxes between 0 and 100 %.

    Args:
        series: Data with a pandas.DatetimeIndex.
        num_boxes: Number of boxes.

    Returns:
        dict: A list with the number of points in each box in keyword "data".
    """
    max_val = series.max()
    min_val = series.min()
    delta = (max_val - min_val) / float(num_boxes)
    eps = 1.0e-6

    boxes = []
    low_limit = min_val
    for idx in range(num_boxes):
        if idx == 0:
            num_value_in_box = ((series >= low_limit * (1 - eps)) & (series <= low_limit + delta * (1 + eps))).sum()
        else:
            num_value_in_box = ((series > low_limit) & (series <= low_limit + delta * (1 + eps))).sum()
        boxes.append(num_value_in_box)
        low_limit += delta

    return {"data": boxes}


def functional_mean(function: Callable, x_vals: List) -> np.ndarray:
    """Convenience method to calculate the mean of a function between each x value.

    The last point is the function called at the last x value. This is required for the
    input and output lists to have the same lengths.

    Args:
        function: Function used to calculate the mean.
        x_vals: Data values used to calculate the mean of a function.

    Raises:
        RuntimeError: Checks that x_vals is not empty.

    Returns:
        numpy.ndarray: Array with the calculated values.
    """
    # Catch empty input
    if len(x_vals) == 0:
        raise UserValueError("No data in the input timeseries.")

    y_vals = np.zeros(len(x_vals))
    for i in range(len(x_vals) - 1):
        y_vals[i] = scipy.integrate.quad(function, x_vals[i], x_vals[i + 1])[0] / (x_vals[i + 1] - x_vals[i])
    y_vals[len(x_vals) - 1] = function(x_vals[-1])
    return y_vals


def is_na_all(data: Union[pd.DataFrame, pd.Series]) -> bool:
    """Convenience method to test if all values in dataframe/series are Nan.

    Args:
        data: Input data.

    Raises:
        UserValueError: Checks that the data type for the input data is pandas.DataFrame or pandas.Series.

    Returns:
        bool: True if all values are Nan, False if all values are not Nan.
    """
    if isinstance(data, pd.Series):
        return data.isnull().all()
    elif isinstance(data, pd.DataFrame):
        return pd.isna(data).all(axis=None)
    else:
        raise UserValueError("Convenience method only supports Series or DataFrame.")


def gaps_detector(timestamps: np.ndarray, threshold: int = 86400) -> np.ndarray:
    """Detect gaps bigger than a threshold in time series.

    Taken from:
    https://github.com/cognitedata/data-quality-
    functions/blob/master/functions/completeness/function/completeness.py.

    Args:
        timestamps: Time array_like.
        threshold: Threshold for gaps. Defaults to 86400 seconds, i.e one full day in epoch time.

    Returns:
        numpy.ndarray: 2D array of begin and end time of gaps in seconds epoch time.
    """
    if len(timestamps) == 0:
        return None

    mask = np.diff(timestamps) >= threshold
    gaps = np.column_stack([timestamps[:-1][mask], timestamps[1:][mask]])

    return gaps


def time_difference(data: pd.Series, unit: str = "ms") -> pd.Series:
    """Calculate the time difference between two consecutive points.

    Args:
        data: Input time series.
        unit: Time unit of uniformity check. Follows Numpy DateTime Units convention. Defaults to "ns".

    Returns:
        pd.Series: Time difference between two consecutive points.

    """
    return data.index.to_series().diff().dropna().astype(dtype=f"timedelta64[{unit}]")


@check_types
def mad(
    data: pd.Series,
) -> float:
    """Median absolute deviation (MAD).

    Median absolute deviation computed for a time series.

    Args:
        data: Time series.

    Returns:
        Median absolute deviation
    """
    return (data - data.median()).abs().median()


def check_uniform(data: pd.DataFrame, unit: str = "ns") -> bool:
    """Convenience method to verify input time series is uniform.

    Args:
        data: Time series to check. Object must have a datetime-like index.
        unit: Time unit of uniformity check. Follows Numpy DateTime Units convention. Defaults to "ns".

    Returns:
        bool: True if input timeseries is uniform, False if not uniform.
    """
    # Get time difference between successive observations
    timediff = np.diff(data.index)
    # Round to desired time unit (e.g. uniform on second scale)
    floored = timediff.astype(dtype=f"timedelta64[{unit}]")
    # Check whether uniform
    return np.all(floored[0] == floored[1:])


def make_uniform(data: pd.DataFrame, resolution: str = "1s", interpolation: str = None) -> pd.DataFrame:
    """Make time series uniform.

    Convenience method to make an input time series dataframe uniform per specified temporal resolution.
    Object must have a datetime-like index.

    Args:
        resolution: Temporal resolution of uniform time series. Follows Pandas DateTime convention. Defaults to "1s".
        interpolation: Method for optional interpolation. Follows pandas Resampler.interpolate. Defaults to None.

    Returns:
        pandas.DataFrame: Uniform Timeseries DataFrame.
    """
    # If no interpolation specified, do not perform interpolation
    if not interpolation:
        data = data.resample(resolution).mean()
    # Otherwise perform desired interpolation
    else:
        data = data.resample(rule=resolution).mean().interpolate(method=interpolation)
    return data


def time_parse(time_window: str = "60min", function_name: str = "") -> pd.Timedelta:
    """Check that time_window is given in a correct format.

    Args:
        time_window: Time window input. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
            If the user gives a number without unit (such as '60'), it will be considered as the number of minutes.
        function_name: Prints the name of the function that called time_parse in the error message if an error is raised. Defaults to "".

    Raises:
        UserValueError: Raise a value error if time_window is given in a wrong format.

    Returns:
        str: parsed time_window in string format.
    """
    if str.isdigit(time_window):  # if user gives '60' it will be considered as '60min'
        warnings.warn(
            f"Missing time unit in argument 'time_window' in {function_name} function, assuming {time_window}min.",
            UserWarning,
        )
        time_window = str(time_window) + "min"
    try:
        return pd.Timedelta(time_window)
    except ValueError:
        raise UserValueError(
            f"Time window should be a string in weeks, days, hours or minutes format:'3w', '10d', '5h', '30min', '10s' not {time_window}."
        ) from None


def time_to_points(data: pd.Series, time_window="60min") -> int:
    """Find the average number of points given a time window.

    Args:
        data: Data with a pandas.DatetimeIndex.
        time_window: Length of the time period to compute the average of points. Defaults to "60min".
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.

    Returns:
        int: average number of points in a time window.
    """
    time_window = time_parse(time_window)
    num_points = data.rolling(time_window).count().mean()

    return round(num_points)


def fill_gaps(
    data: pd.Series,
    granularity: Optional[pd.Timedelta] = None,
    ffill_resolution: Optional[pd.Timedelta] = None,
    interpolate_resolution: Optional[pd.Timedelta] = None,
    method: Literal["ffill", "backfill", "bfill", "pad"] = "ffill",
) -> pd.Series:
    """Mask nan/fill gaps based on the time size of missing data.

    Args:
        data: Input time series. Must be uniform.
        granularity: Temporal resolution of uniform time series, before resampling. Defaults to None. If not given as input, it will be inferred from the data.
        ffill_resolution: Gaps smaller than this will be filled with method. Defaults to None.
        interpolate_resolution: Gaps smaller than this will be interpolated, larger than this will be filled by noise. Defaults to None.
        method: Method used to fill nan. Defaults to 'ffill', but ‘backfill’, ‘bfill’ and ‘pad’ can also be used.

    Returns:
        pandas.DataFrame or pd.Series: Uniform, time series without nan values.

    Raises:
        UserRuntimeError: The input time series is not uniform
        UserValueError: granularity can not be 0
        UserValueError: ffill_resolution can not be 0
        UserValueError: interpolate_resolution can not be 0
    """
    if not check_uniform(data):
        raise UserRuntimeError("The input time series is not uniform")

    if not data.isnull().values.any():
        # no need to proceed
        return data

    if granularity is None:
        # If granularity is not given as input. Granularity will be inferred from the data
        granularity = pd.Timedelta(data.index.inferred_freq)
    if granularity == pd.Timedelta("0T"):
        raise UserValueError("granularity can not be 0")

    if ffill_resolution == pd.Timedelta("0T"):
        raise UserValueError("ffill_resolution can not be 0")

    if interpolate_resolution == pd.Timedelta("0T"):
        raise UserValueError("interpolate_resolution can not be 0")

    small_gap = 0
    inter_gap = None
    mask_gap = 0

    if ffill_resolution is not None:
        small_gap = granularity / ffill_resolution
    if interpolate_resolution is not None:
        mask_gap = granularity / interpolate_resolution
        # inter_gap has to be above 0, but can be None
        inter_gap = max(1, int(mask_gap))

    # ffill all, after interpolate and adding noise
    # create groups based on size of gap
    nan_cumsum = (
        data.isnull().astype("int64").groupby(data.notnull().astype("int64").cumsum()).cumsum().mask(lambda x: x == 0)
    )
    helper = data.notnull()
    nan_group_size = nan_cumsum.groupby(helper.cumsum()).transform("max").mask(helper)

    # start with random filling for the big gaps
    mask = nan_group_size > inter_gap
    n = len(data[mask])
    data[mask] = np.random.normal(loc=np.mean(data), scale=np.std(data), size=n)

    # interpolate the inteprolate range
    mask = np.logical_and(nan_group_size > small_gap, nan_group_size <= mask_gap)
    data[mask] = data.interpolate(method="linear", limit_direction="forward", limit=inter_gap)

    return data.fillna(method=method)


def number_of_events(out: pd.Series):
    """Calculate the number of events detected from a step time series."""
    return len(np.where(out.values == 1)[0]) / 2


@check_types
def scalar_to_pandas_series(data: Union[float, pd.Series]) -> pd.Series:
    """Convert data to series if it is not already.

    Args:
        data: Data that needs to be series.

    Returns:
        pd.Series: Data as series.
    """
    if isinstance(data, float):
        # IDEA: Use const_value function instead, but have to avoid circular import
        # Needs at least two values covering the whole time span of interest to work with auto_align
        data = pd.Series(data, index=pd.date_range(start="1970", end=pd.Timestamp.now(), periods=2))
    return data



# Copyright 2022 Cognite AS
from __future__ import annotations

from typing import List, Optional, Tuple

import pandas as pd

from indsl.data_quality import uncertainty_rstd

from ..base import DataQualityScore, DataQualityScoreAnalyser


class UncertaintyDataQualityScoreAnalyser(DataQualityScoreAnalyser):
    """Uncertainty data quality score analyser class."""

    @staticmethod
    def _limit_series_to_analysis_period(
        series: pd.Series, analysis_start: pd.Timestamp, analysis_end: pd.Timestamp
    ) -> pd.Series:
        return series[(analysis_start <= series.index) & (series.index <= analysis_end)]

    def compute_score(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        threshold=0.02,
        resample_rate: pd.Timedelta = pd.Timedelta("30m"),
        emd_sift_thresh: float = 1e-8,
        emd_max_num_imfs: Optional[int] = None,
        emd_error_tolerance: float = 0.05,
    ) -> DataQualityScore:
        """Compute uncertainty data quality score.

        The score measures events where the uncertainty of the time series values is above the user-specified threshold.
        The uncertainty is estimated using the uncertainty_rstd function, the segment size is controlled by the resample_rate parameter.

        Args:
            analysis_start: Analyis start time
            analysis_end: Analyis end time
            threshold (float): Threshold
                Uncertainty values above this threshold are counted as events in the data quality score. Default: 0.02
            resample_rate: Window aggregate
                Perform the uncertainty analysis on aggregate data. Default: 30 minutes
            emd_sift_thresh: Sifting threshold
                Threshold to stop EMD sifting process. This threshold is based on the Cauchy convergence test and represent the
                residue between two consecutive oscillatory components (IMFs). A small threshold (close to zero) will
                result in more components extracted. Typically, a few IMFs are enough to build the main trend. Choosing a
                high threshold might not affect the outcome. Defaults to 1e-8.
            emd_max_num_imfs: Maximum number of components
                Maximum number of EMD oscillatory components (or IMFs) to estimate the main trend. If no value (None) is defined
                the process continues until sifting threshold is reached. Defaults to None.
            emd_error_tolerance: Energy tolerance
                Threshold for the EMD cross energy ratio validation used for choosing oscillatory components or IMFs. Defaults to
                0.05.

        Returns:
            DataQualityScore: Data quality score
        """
        if len(self.series) < 10:
            return DataQualityScore(analysis_start, analysis_end, [])

        relative_std_deviation = uncertainty_rstd(
            self.series, resample_rate, emd_sift_thresh, emd_max_num_imfs, emd_error_tolerance
        )

        event_marker_resampler = (relative_std_deviation > threshold).resample(resample_rate, origin="start")
        events_series = event_marker_resampler.sum() / event_marker_resampler.count()
        events_series = events_series[events_series > 0]
        events = self._convert_marker_series_to_event_list(events_series, resample_rate, analysis_start, analysis_end)
        score = DataQualityScore(analysis_start, analysis_end, events)

        return score

    @staticmethod
    def _plot(series, estimated_truth, events, score):  # pragma: no cover
        from matplotlib import pyplot as plt

        _, ax1 = plt.subplots()
        ax1.plot(series, label="Signal with noise")
        ax1.plot(estimated_truth, label="Estimated truth")

        for start, stop in events:
            plt.axvspan(start, stop, color="red", alpha=0.5, zorder=10)

        plt.title(f"Uncertainty score: {score.data_quality_score:.4f}")

        plt.show()

    @staticmethod
    def _convert_marker_series_to_event_list(
        events_series: pd.Series,
        window_aggregate: pd.Timedelta,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
    ) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:

        # Create initial list of events
        def limit_to_analysis_period(timestamp):
            return min(max(timestamp, analysis_start), analysis_end)

        events = [
            (limit_to_analysis_period(event_start), limit_to_analysis_period(event_start + window_aggregate))
            for event_start in events_series.index
        ]

        # Remove empty events
        events = [(start, end) for start, end in events if end - start > pd.Timedelta(0)]

        # Merge contiguous events
        merged_events: List[List[pd.Timestamp]] = []
        for event_start, event_end in events:

            if len(merged_events) > 0 and event_start == merged_events[-1][1]:
                merged_events[-1][1] = event_end
            else:
                merged_events.append([event_start, event_end])

        return [tuple(event) for event in merged_events]  # type: ignore



# Copyright 2022 Cognite AS
from typing import Literal, Optional

import numpy as np
import pandas as pd

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index, validate_series_is_not_empty


class Cusum:
    """Cumulative sum (CUSUM).

    This technique calculates the cumulative sum of positive and negative changes (g+t and g−t) in the data (x) and compares them to a threshold.
    When this threshold is exceeded, a change is detected (time_talarm), and the cumulative sum restarts from zero.
    To avoid the detection of a change in absence of an actual change or a slow drift, this algorithm also depends on a parameter drift for drift correction.
    Remove extreme standalone outliers before using this technique to get better results.

    Attributes:
        data (pd.Series): Time series data.
        mean_data (pd.Series): Smoothed time series.
        threshold (float): Amplitid threshold.
        drift (float): Drift term.
        detect (str): Type of changes to detect.
        predict_ending (bool): Predict end point.
        alpha (float): Smoothing factor for mean data.
        time_alarm (np.array): Detected change points.
        time_alarm_initial (np.array): Start of the changes.
        time_alarm_final (np.array): End of the changes.
        cumulative_positive_sum (np.array): Cumulative sum of positive changes.
        cumulative_negative_sum (np.array): Cumulative sum of negative changes.

    References:
        https://nbviewer.org/github/demotu/detecta/blob/master/docs/detect_cusum.ipynb
    """

    @check_types
    def __init__(
        self,
        data: pd.Series,
        threshold: Optional[float] = None,
        drift: Optional[float] = None,
        detect: Literal["both", "increase", "decrease"] = "both",
        predict_ending: bool = True,
        alpha: float = 0.05,
    ):
        """Cusum init function.

        Args:
            data: Time series
            threshold: Amplitid threshold.
                Cumulative changes are compared to this threshold. Defaults to None.
                When this threshold is exceeded, a change is detected, and the cumulative sum restarts from 0.
                If the threshold is not provided, it is assigned to 5 * standard_deviation of the data.
            drift: Drift term.
                Prevents any change in the absence of change. Defaults to None.
                If fewer false alarms are wanted, try to increase `drift`.
                If the threshold is not provided, it is assigned to (2 * data_standard_deviation - data_mean) / 2.
            detect: Type of changes to detect:
                    * "both" for detecting both increasing and decreasing changes in the data (default)
                    * "increase" for detecting increasing changes in the data
                    * "decrease" for detecting decreasing changes in the data
            predict_ending: Predict end point.
                Prolongs the change until the predicted end point. Defaults to True.
                If false, single change points are detected.
            alpha: Smoothing factor.
                Value between 0 < alpha <= 1. Defaults to 0.05.

        Raises:
            UserTypeError: If a time series with the wrong index is provided.
            UserValueError: If an empty time series is passed into the function.
        """
        self.data = data
        self._validate_time_series()

        # user inputs
        self.threshold = 5 * data.std() if not threshold else threshold
        self.drift = (2 * data.std() - data.mean()) / 2 if not drift else drift
        self.detect = detect
        self.predict_ending = predict_ending
        self.alpha = alpha

        # smooth time series used as mean data
        self.mean_data = data.ewm(alpha=alpha).mean()

        # store results
        self.time_alarm = np.array([])
        self.time_alarm_initial = np.array([])
        self.time_alarm_final = np.array([])
        self.cumulative_positive_sum = np.array([])
        self.cumulative_negative_sum = np.array([])

    def _validate_time_series(self):
        """Validate time series.

        Raises:
            UserValueError: If an empty time series is passed into the function.
            UserValueError: If a time series with the wrong index is provided.
        """
        validate_series_is_not_empty(self.data)
        validate_series_has_time_index(self.data)

        if not self.data.index.is_monotonic_increasing:
            raise UserValueError("Time series index is not increasing.")

    @check_types
    def _detect_changes(self, endings: bool = False):
        """Detection of change points.

        Args:
            endings: Detect only the end of the changes previously detected.
                Defaults to False.
        """
        if endings:  # revert the datapoints
            x = self.data.values[::-1]
            m = self.mean_data.values[::-1]

        else:
            x = self.data.values
            m = self.mean_data.values

        cumulative_positive, cumulative_negative = np.zeros(x.size), np.zeros(x.size)
        time_alarm, time_alarm_initial, time_alarm_final = (
            np.empty(0, dtype=np.int64),
            np.empty(0, dtype=np.int64),
            np.empty(0, dtype=np.int64),
        )
        time_alarm_positive, time_alarm_negative = 0, 0
        s = x - m

        for i in range(1, x.size):
            cumulative_positive[i] = cumulative_positive[i - 1] + s[i] - self.drift  # cumulative sum for + change
            cumulative_negative[i] = cumulative_negative[i - 1] - s[i] - self.drift  # cumulative sum for - change

            if cumulative_positive[i] < 0:
                cumulative_positive[i], time_alarm_positive = 0, i
            if cumulative_negative[i] < 0:
                cumulative_negative[i], time_alarm_negative = 0, i

            # detect changes
            if self.detect == "increase" and (cumulative_positive[i] > self.threshold):
                time_alarm = np.append(time_alarm, i)  # alarm index for a positive change
                time_alarm_initial = np.append(time_alarm_initial, time_alarm_positive)  # start
                cumulative_positive[i], cumulative_negative[i] = 0, 0  # reset alarm
            elif self.detect == "decrease" and (cumulative_negative[i] > self.threshold):
                time_alarm = np.append(time_alarm, i)  # alarm index for a negative change
                time_alarm_initial = np.append(time_alarm_initial, time_alarm_negative)  # start
                cumulative_positive[i], cumulative_negative[i] = 0, 0  # reset alarm
            elif self.detect == "both" and (
                cumulative_positive[i] > self.threshold or cumulative_negative[i] > self.threshold
            ):
                time_alarm = np.append(time_alarm, i)  # alarm index for positive or negative change
                time_alarm_initial = np.append(
                    time_alarm_initial,
                    time_alarm_positive if cumulative_positive[i] > self.threshold else time_alarm_negative,
                )  # start
                cumulative_positive[i], cumulative_negative[i] = 0, 0  # reset alarm

        # Eliminate repeated changes, changes that have the same beginning
        time_alarm_initial, ind = np.unique(time_alarm_initial, return_index=True)
        time_alarm = time_alarm[ind]

        # store results as attributes
        if not endings:
            self.time_alarm = time_alarm
            self.time_alarm_initial = time_alarm_initial
            self.time_alarm_final = time_alarm_final

            self.cumulative_positive_sum = pd.Series(cumulative_positive, index=self.data.index)
            self.cumulative_negative_sum = pd.Series(cumulative_negative, index=self.data.index)

        else:
            self.time_alarm_final = self.data.values.size - time_alarm_initial[::-1] - 1

    def _detect_ending(self):
        """Estimation of when the change ends (offline form)."""
        self._detect_changes(endings=True)

        self._fix_time_alarm_size_errors()

    def _fix_time_alarm_size_errors(self):
        if self.time_alarm_initial.size != self.time_alarm_final.size:
            if self.time_alarm_initial.size < self.time_alarm_final.size:
                self.time_alarm_final = self.time_alarm_final[
                    [np.argmax(self.time_alarm_final >= i) for i in self.time_alarm]
                ]
            else:
                ind: np.array = [np.argmax(i >= self.time_alarm[::-1]) - 1 for i in self.time_alarm_final]
                self.time_alarm = self.time_alarm[ind]
                self.time_alarm_initial = self.time_alarm_initial[ind]
        # Delete intercalated changes (the ending of the change is after
        # the beginning of the next change)
        ind = self.time_alarm_final[:-1] - self.time_alarm_initial[1:] > 0
        if ind.any():
            self.time_alarm = self.time_alarm[~np.append(False, ind)]
            self.time_alarm_initial = self.time_alarm_initial[~np.append(False, ind)]
            self.time_alarm_final = self.time_alarm_final[~np.append(ind, False)]

    def _plot_cusum(self) -> None:
        """Plot cusum results.

        Plots 2 figures containing the results of the cusum function:
        1. Raw data, smoothed data, detected change points, start of the change and end of the change (if predict_ending = True).
        2. Cumulative sums for increasing and decreasing changes.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("matplotlib is not available.")
        else:
            _, (ax1, ax2) = plt.subplots(2, 1, figsize=(15, 10))
            time_alarm_initial_index = [pd.Timestamp(self.data.index[tai_val]) for tai_val in self.time_alarm_initial]
            time_alarm_final_index = [pd.Timestamp(self.data.index[taf_val]) for taf_val in self.time_alarm_final]
            time_alarm_index = [pd.Timestamp(self.data.index[ta_val]) for ta_val in self.time_alarm]

            ax1.plot(self.data.index, self.data.values, "-go", lw=2, label="Raw data")
            ax1.plot(self.data.index, self.mean_data.values, "-o", color="orange", lw=2, label="Smoothed data")

            if len(self.time_alarm):
                ax1.plot(
                    time_alarm_initial_index,
                    self.data[self.time_alarm_initial],
                    ">",
                    mfc="b",
                    mec="g",
                    ms=10,
                    label="Start",
                )
                if self.predict_ending:
                    ax1.plot(
                        time_alarm_final_index,
                        self.data.values[self.time_alarm_final],
                        "<",
                        mfc="b",
                        mec="g",
                        ms=10,
                        label="Ending",
                    )
                ax1.plot(
                    time_alarm_index,
                    self.data.values[self.time_alarm],
                    "o",
                    mfc="r",
                    mec="m",
                    mew=1,
                    ms=10,
                    label="Alarm",
                )
                ax1.legend(loc="best", framealpha=0.5, numpoints=1)

            ymin, ymax = (
                self.data.values[np.isfinite(self.data.values)].min(),
                self.data.values[np.isfinite(self.data.values)].max(),
            )
            yrange = ymax - ymin if ymax > ymin else 1
            ax1.set_ylim(ymin - 0.1 * yrange, ymax + 0.1 * yrange)
            ax1.set_title(
                f"Detected changes in the data (threshold= {self.threshold}, drift= {self.drift}): N changes ={len(self.time_alarm_initial)}"
            )
            ax2.plot(self.data.index, self.cumulative_positive_sum, "-oy", label="Increasing changes", lw=0.8)
            ax2.plot(self.data.index, self.cumulative_negative_sum, "-om", label="Decreasing changes", lw=0.8)

            # ax2.set_ylim(-0.01*threshold, 1.1*threshold)
            ax2.axhline(self.threshold, color="r", label=f"Threshold at {self.threshold}")
            ax2.set_title("Cumulative sums of increasing and decreasing changes.")
            ax2.legend(loc="best", framealpha=0.5, numpoints=1)
            plt.tight_layout()
            plt.show()

    def _create_binary_output(self) -> pd.Series:
        """Binary pd.Series with the detected changes.

        Returns:
            pd.Series: Binary time series.
            Change detected = 1, No change detected = 0.
        """
        output_series: pd.Series = pd.Series([0] * self.data.size, index=self.data.index)

        if len(self.time_alarm_final) > 0:
            # amp = data[taf] - data[tai]
            for i in range(len(self.time_alarm_initial)):
                output_series[self.time_alarm_initial[i] : self.time_alarm_final[i] + 1] = 1
        else:
            output_series[self.time_alarm_initial] = 1

        return output_series

    @check_types
    def cusum(
        self,
        plot_cusum: bool = False,
    ) -> pd.Series:
        """Cumulative sum (CUSUM).

        This technique calculates the cumulative sum of positive and negative changes (g+t and g−t) in the data (x) and compares them to a threshold.
        When this threshold is exceeded a change is detected (ttalarm) and the cumulative sum restarts from zero.
        To avoid the detection of a change in absence of an actual change or a slow drift, this algorithm also depends on a parameter drift for drift correction.
        To get better results, it is recommended to remove extreme standalone outliers before using this technique.

        Returns:
            pd.Series: Binary time series.
            Change detected = 1, No change detected = 0.

        Raises:
            UserTypeError: If a time series with the wrong index is provided.
            UserValueError: If an empty time series is passed into the function.

        References:
            https://nbviewer.org/github/demotu/detecta/blob/master/docs/detect_cusum.ipynb
        """
        self._detect_changes()

        if self.time_alarm_initial.size and self.predict_ending:
            self._detect_ending()
        if plot_cusum:
            self._plot_cusum()
        return self._create_binary_output()


@check_types
def cusum(
    data: pd.Series,
    threshold: Optional[float] = None,
    drift: Optional[float] = None,
    detect: Literal["both", "increase", "decrease"] = "both",
    predict_ending: bool = True,
    alpha: float = 0.05,
    return_series_type: Literal[
        "cusum_binary_result", "mean_data", "positive_cumulative_sum", "negative_cumulative_sum"
    ] = "cusum_binary_result",
) -> pd.Series:
    """Cumulative sum (CUSUM).

    This technique calculates the cumulative sum of positive and negative changes (g+t and g−t) in the data (x) and compares them to a threshold.
    When this threshold is exceeded, a change is detected (ttalarm), and the cumulative sum restarts from zero.
    To avoid the detection of a change in absence of an actual change or a slow drift, this algorithm also depends on a parameter `drift` for drift correction.
    Remove extreme standalone outliers before using this technique to get a better result.

    Typical uses of this function:

    1. Set the type of series to return to "mean_data" to visualize the smoothed data. Leave the rest of the parameters to their default values.
    2. Adjust the alpha parameter to get the desired smoothing for the data.
    3. Set the type of series to return to "positive_cumulative_sum" or "negative_cumulative_sum" to visualize the cumulative sum of the positive or negative changes.
    4. Adjust the threshold and drift accordingly to get the desired number of change points.
    5. Set the type of series to return to "cusum_binary_result" to visualize the detected changes.

    Args:
        data: Time series.
        threshold: Amplitude threshold.
            Cumulative changes are compared to this threshold. Defaults to None.
            When this is exceeded a change is detected and the cumulative sum restarts from 0.
            If the threshold is not provided, it is assigned to 5 * standard_deviation of the data.
        drift: Drift term.
            Prevents any change in the absence of change. Defaults to None.
            If fewer false alarms are wanted, try to increase `drift`.
            If the threshold is not provided, it is assigned to (2 * data_standard_deviation - data_mean) / 2.
        detect: Type of changes to detect.
              Options are:
                * "both" for detecting both increasing and decreasing changes in the data (default)
                * "increase" for detecting increasing changes in the data
                * "decrease" for detecting decreasing changes in the data
        predict_ending: Predict end point.
            Prolongs the change until the predicted end point. Defaults to True.
            If false, single change points are detected.
        alpha: Smoothing factor.
            Value between 0 < alpha <= 1. Defaults to 0.05.
        return_series_type: Type of series to return.
            Defaults to "cusum_binary_result".
            This option allows the user to visualize the intermediate steps of the algorithm.
            Options are:

            * "cusum_binary_result" returns the cusum results as a binary time series. Change detected = 1, No change detected = 0.
            * "mean_data" returns the smoothed data.
            * "positive_cumulative_sum" returns the positive cumulative sum.
            * "negative_cumulative_sum" returns the negative cumulative sum.

    Returns:
        pd.Series: Time series.
            Specified in the return_series_type parameter.

    Raises:
        UserTypeError: If a time series with the wrong index is provided.
        UserValueError: If an empty time series is passed into the function.

    References:
        https://nbviewer.org/github/demotu/detecta/blob/master/docs/detect_cusum.ipynb
    """
    cusum_detection = Cusum(
        data=data,
        threshold=threshold,
        drift=drift,
        detect=detect,
        predict_ending=predict_ending,
        alpha=alpha,
    )

    if return_series_type == "mean_data":
        return cusum_detection.mean_data
    else:
        cusum_binary_result = cusum_detection.cusum(plot_cusum=False)

        if return_series_type == "positive_cumulative_sum":
            return cusum_detection.cumulative_positive_sum
        elif return_series_type == "negative_cumulative_sum":
            return cusum_detection.cumulative_negative_sum
        return cusum_binary_result



# Copyright 2022 Cognite AS
from typing import Callable, List

import numpy as np
import pandas as pd

from numba import jit

from indsl.detect.utils import resample_timeseries

from ..exceptions import UserRuntimeError, UserTypeError, UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index


@check_types
def cpd_ed_pelt(data: pd.Series, min_distance: int = 1) -> pd.Series:
    """Change Point Detection.

    This function detects change points in a time series. The time series is split into "statistically homogeneous" segments using the
    ED Pelt change point detection algorithm while observing the minimum distance argument.

    Args:
        data: Time series

        min_distance: Minimum distance.
            Specifies the minimum point wise distance for each segment that will be considered in the Change
            Point Detection algorithm.

    Returns:
        pandas.Series: Time series.
        Binary time series.
    """
    validate_series_has_time_index(data)

    data_resampled: pd.Series = resample_timeseries(data=data)

    # the maximum allowable distance is half the number of datapoints
    max_distance: int = int(np.floor(len(data_resampled) / 2))
    if min_distance > max_distance:
        raise UserRuntimeError(
            f"Minimum segment distance argument ({min_distance}) is larger than the maximum allowable segment distance ({max_distance})"
        )

    change_points: np.ndarray = ed_pelt(data=data_resampled.values, min_distance=min_distance)

    # create the binary time series full of zeros
    change_points_ts: pd.Series = pd.Series(index=data_resampled.index, data=[0] * len(data_resampled))

    # add the change points to the time series
    for cp in change_points:
        change_points_ts[cp] = 1
        # add the value of 0 to 1ns before and after the timestamp of the current change point
        change_points_ts = pd.concat(
            [
                change_points_ts,
                pd.Series(
                    index=[
                        data_resampled.index[cp] - pd.Timedelta(value=1, unit="nanoseconds"),
                        data_resampled.index[cp] + pd.Timedelta(value=1, unit="nanoseconds"),
                    ],
                    data=[0, 0],
                ),
            ]
        )

    # returns a time series with values at the change points and nan's everywhere else
    # this will be rendered as vertical lines in the change point locations
    return change_points_ts.sort_index()


@check_types
def get_partial_sums(data: np.ndarray, k: int) -> np.ndarray:
    """Partial sums.

    Partial sums for empirical CDF (formula (2.1) from Section 2.1 "Model" in [Haynes2017])

    Args:
        data: Input data
        k: Number of quantiles

    Returns:
        np.ndarray: Partial Sums
    """
    n: int = len(data)
    partial_sums: np.ndarray = np.zeros(shape=(k, n + 1), dtype=int)
    sorted_data: np.ndarray = np.sort(data)

    for i in range(k):
        z: float = -1 + (2 * i + 1.0) / k  # values from (-1 + 1 / k) to (1 - 1 / k) with step = 2 / k
        p: float = 1.0 / (1 + np.power(2 * n - 1, -z))  # values from 0.0 to 1.0
        t: float = sorted_data[int(np.trunc((n - 1) * p))]  # quantile value, formula (2.1) in[Haynes2017]

        for tau in range(1, n + 1):
            partial_sums[i, tau] = partial_sums[i, tau - 1]
            if data[tau - 1] < t:
                partial_sums[i, tau] += 2  # we use doubled value (2) instead of original 1.0
            if data[tau - 1] == t:
                partial_sums[i, tau] += 1  # we use doubled value (1) instead of original 0.5

    return partial_sums


@jit(nopython=True)
def get_segment_cost(partial_sums: np.ndarray, tau1: int, tau2: int, k: int, n: int) -> float:
    """Calculates the cost of the (tau1; tau2] segment.

    Args:
        partial_sums: Partial sums for empirical CDF
        tau1: Start of the segment
        tau2: End of the segment
        k: Number of quantiles
        n: Number of datapoints

    Returns:
        float: Segment cost
    """
    sum: float = 0
    for i in range(k):
        # actual_sum is (count(data[j] < t) * 2 + count(data[j] == t) * 1) for j=tau1..tau2-1
        actual_sum: int = partial_sums[i, tau2] - partial_sums[i, tau1]

        # we skip these two cases (correspond to fit = 0 or fit = 1) because of invalid np.log values
        if (actual_sum != 0) & (actual_sum != (tau2 - tau1) * 2):
            # empirical CDF F_i(t) (Section 2.1 "Model" in [Haynes2017])
            fit: float = actual_sum * 0.5 / (tau2 - tau1)
            # segment cost L_np (Section 2.2 "Nonparametric maximum likelihood" in [Haynes2017])
            lnp: float = (tau2 - tau1) * (fit * np.log(fit) + (1 - fit) * np.log(1 - fit))
            sum += lnp

    c: float = -np.log(2 * n - 1)  # Constant from Lemma 3.1 in [Haynes2017]
    return 2.0 * c / k * sum  # See Section 3.1 "Discrete approximation" in [Haynes2017]


@check_types
def ed_pelt(data: np.ndarray, min_distance: int = 1) -> np.ndarray:
    """The ED-PELT algorithm for change point detection.

    For a given array of `float` values, this algorithm detects locations of change points that splits original series of values into
    "statistically homogeneous" segments. Such points correspond to moments when statistical properties of the
    distribution are changing. This method supports nonparametric distributions and has O(N*log(N)) algorithmic
    complexity.

    Args:
        data: An array with sensor values.
        min_distance: Minimum distance between changepoints. Defaults to 1.

    Raises:
        UserTypeError: data needs to be a numpy array containing float elements.
        UserTypeError: min_distance needs to be a value between 1 and len(data).

    Returns:
        np.ndarray: Returns an array with 1-based indexes of changepoint. Changepoints correspond to the end of the
        detected segments.

    References:
        1. [Haynes2017] Haynes, Kaylea, Paul Fearnhead, and Idris A. Eckley. "A computationally efficient nonparametric
        approach for changepoint detection." Statistics and Computing 27, no. 5 (2017): 1293-1305.
        https://doi.org/10.1007/s11222-016-9687-5

        2. [Killick2012] Killick, Rebecca, Paul Fearnhead, and Idris A. Eckley. "Optimal detection of changepoints with
        a linear computational cost." Journal of the American Statistical Association 107, no. 500 (2012): 1590-1598.
        https://arxiv.org/pdf/1101.1438.pdf

    Based on the ED-Pelt C# implementation from (c) 2019 Andrey Akinshin
    Licensed under The MIT License https://opensource.org/licenses/MIT
    """
    # check if the provided data is in the correct format
    if data.dtype.kind not in np.typecodes["AllFloat"]:
        raise UserTypeError("data argument need to be a numpy array containing float elements")

    # we will use `n` as the number of elements in the `data` array
    n: int = len(data)

    # check corner cases
    if n <= 2:
        return np.array([])
    if not 1 <= min_distance <= n:
        raise UserValueError(f"min_distance ({min_distance}) should be in range from 1 to {n}")

    # the penalty which we add to the final cost for each additional changepoint
    # here we use the Modified Bayesian Information Criterion
    penalty: float = 3 * np.log(n)

    # `k` is the number of quantiles that we use to approximate an integral during the segment cost evaluation
    # we use `k = ceiling(4 * log(n))` as suggested in the Section 4.3 "Choice of K in ED-PELT" in [Haynes2017]
    # `k` can't be greater than `n`, so we should always use the `Min` function here (important for n <= 8)
    k: int = min(n, int(np.ceil(4 * np.log(n))))

    # we should precalculate sums for empirical CDF, it will allow fast evaluating of the segment cost
    partial_sums: np.ndarray = get_partial_sums(data, k)

    # since we use the same values of `partial_sums`, `k`, `n` all the time,
    # we introduce a shortcut `cost(tau1, tau2)` for segment cost evaluation.
    # hereinafter, we use `tau` to name variables that are changepoint candidates.
    cost: Callable[[int, int], float] = lambda tau1, tau2: get_segment_cost(partial_sums, tau1, tau2, k, n)

    # we will use dynamic programming to find the best solution; `best_cost` is the cost array.
    # `best_cost[i]` is the cost for subarray `data[0..i - 1]`.
    # it's a 1-based array (`data[0]`..`data[n-1]` correspond to `best_cost[1]`..`best_cost[n]`)
    best_cost: np.ndarray = np.zeros(n + 1)
    best_cost[0] = -penalty
    for current_tau in range(min_distance, 2 * min_distance):
        best_cost[current_tau] = cost(0, current_tau)

    # `previous_change_point_index` is an array of references to previous changepoints. If the current segment ends at
    # the position `i`, the previous segment ends at the position `previous_change_point_index[i]`. It's a 1-based
    # array(`data[0]`..`data[n - 1]` correspond to the `previous_change_point_index[1]`..
    # `previous_change_point_index[n]`)
    previous_change_point_index: np.ndarray = np.zeros(shape=n + 1, dtype=int)

    # we use PELT (Pruned Exact Linear Time) approach which means that instead of enumerating all possible previous
    # tau values, we use a whitelist of "good" tau values that can be used in the optimal solution. If we are 100%
    # sure that some of the tau values will not help us to form the optimal solution, such values should be
    # removed. See [Killick2012] for details.
    previous_taus: List[int] = [0, min_distance]

    # following the dynamic programming approach, we enumerate all tau positions. For each `current_tau`, we pretend
    # that it's the end of the last segment and trying to find the end of the previous segment.
    for current_tau in range(2 * min_distance, n + 1):
        # for each previous tau, we should calculate the cost of taking this tau as the end of the previous
        # segment. This cost equals the cost for the `previous_tau` plus cost of the new segment (from `previous_tau`
        # to `current_tau`) plus penalty for the new changepoint.
        cost_for_previous_tau: List[float] = [
            best_cost[previous_tau] + cost(previous_tau, current_tau) + penalty for previous_tau in previous_taus
        ]

        # Now we should choose the tau that provides the minimum possible cost.
        best_previous_tau_index: int = np.argmin(cost_for_previous_tau)
        best_cost[current_tau] = cost_for_previous_tau[best_previous_tau_index]
        previous_change_point_index[current_tau] = previous_taus[best_previous_tau_index]

        # Prune phase: we remove "useless" tau values that will not help to achieve minimum cost in the future
        current_best_cost: float = best_cost[current_tau]
        new_previous_taus_size: int = 0
        for i in range(len(previous_taus)):
            if cost_for_previous_tau[i] < current_best_cost + penalty:
                previous_taus[new_previous_taus_size] = previous_taus[i]
                new_previous_taus_size += 1
        previous_taus = previous_taus[:new_previous_taus_size]

        # We add a new tau value that is located on the `min_distance` distance from the next `current_tau` value
        previous_taus.append(current_tau - (min_distance - 1))

    # here we collect the result list of changepoint indexes `change_point_indexes` using `previous_change_point_index`
    change_point_indexes: List[int] = []
    current_index: int = previous_change_point_index[n]  # The index of the end of the last segment is `n`
    while current_index != 0:
        change_point_indexes.append(current_index)
        current_index = previous_change_point_index[current_index]

    # sort the change_point_indexes and transform it to a numpy ndarray
    result: np.ndarray = np.asarray(change_point_indexes, dtype=int)
    result_sorted: np.ndarray = np.sort(result)

    return result_sorted



# Copyright 2022 Cognite AS
from .interpolate import interpolate
from .reindex import reindex
from .resample import resample, resample_to_granularity


__all__ = [
    "interpolate",
    "resample",
    "resample_to_granularity",
    "reindex",
]


TOOLBOX_NAME = "Resample"



# Copyright 2021 Cognite AS
import numpy as np
import pandas as pd

from scipy.integrate import cumulative_trapezoid

from indsl import versioning
from indsl.exceptions import UserValueError
from indsl.ts_utils.ts_utils import granularity_to_ms
from indsl.type_check import check_types


def granularity_to_ns(granularity, time_unit):  # noqa: D103
    if time_unit is None or time_unit.lower() != "auto":
        granularity = time_unit
    return granularity_to_ms(granularity) * 1e6


@check_types
@versioning.register(version="1.0", deprecated=True)
def trapezoidal_integration(series: pd.Series, granularity: str = "1h", time_unit: str = "auto"):
    """Integration.

    Cumulative integration using trapezoidal rule with an optional user-defined time unit.

    Args:
        series: Time series.
        granularity: Granularity.
            Current granularity for the chart on-screen (auto-given).
        time_unit: Frequency.
            User defined granularity to potentially override unit of time.
            Accepts integer followed by time unit string (s|m|h|d). For example: '1s', '5m', '3h' or '1d'.

    Returns:
        pandas.Series: Cumulative integral.
    """
    if len(series) < 1:
        raise UserValueError(f"Expected series to be of length > 0, got length {len(series)}")
    gran_ns = granularity_to_ns(granularity, time_unit)
    arr = cumulative_trapezoid(series, series.index.view(np.int64) / gran_ns, initial=0.0)
    return pd.Series(arr, index=series.index)


@check_types
@versioning.register(version="1.0", deprecated=True)
def differentiate(series: pd.Series, granularity: str = "1h", time_unit: str = "auto"):
    """Differentiation.

    Differentiation (finite difference) using a second-order accurate numerical method (central difference).
    Boundary points are computed using a first-order accurate method.

    Args:
        series: Time series.
        granularity: Granularity.
            Current granularity for the chart on-screen (auto-given).
        time_unit: Frequency.
            User defined granularity to potentially override unit of time.
            Accepts integer followed by time unit string (s|m|h|d). For example: '1s', '5m', '3h' or '1d'.

    Returns:
        pandas.Series: First order derivative.
    """
    if len(series) < 2:
        raise UserValueError(f"Expected series to be of length > 1, got length {len(series)}")
    gran_ns = granularity_to_ns(granularity, time_unit)
    arr = np.gradient(series, series.index.view(np.int64) / gran_ns)
    return pd.Series(arr, index=series.index)



from typing import Union

import numpy as np
import pandas as pd

from indsl.resample.auto_align import auto_align

from ..exceptions import UserValueError
from ..type_check import check_types


@check_types
def _pct_to_fraction(pct_value: Union[float, int]):
    return pct_value / 100.0


@check_types
def calculate_well_prod_status(
    master_valve: pd.Series,
    wing_valve: pd.Series,
    choke_valve: pd.Series,
    threshold_master: float = 1,
    threshold_wing: float = 1,
    threshold_choke: float = 5,
    align_timesteps: bool = False,
) -> pd.Series:
    """Check if the well is producing.

    Determine if the well is producing. In order for this to be the case, the following has to happen:

        * All Master, Wing and Choke data have to come from the same well.
        * Check if the master, wing and choke valve openings are above their respective threshold values at a given time.
        * If any of the valves are below the threshold opening, then the well is closed.
        * If all of the valves are above the threshold opening, then the well is open.
        * Threshold values should be between 0-100.

    Args:
        master_valve: Master Valve
            Time series of the master valve.
        wing_valve:  Wing Valve
            Time series of the wing valve.
        choke_valve:  Choke Valve
            Time series of the choke valve.
        threshold_master: Master threshold
            Threshold percentage value from 0%-100%.
        threshold_wing: Wing threshold
            Threshold percentage value from 0%-100%.
        threshold_choke: Choke threshold
            Threshold percentage value from 0%-100%.
        align_timesteps: Auto-align
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Well Status
            Well production status (1 means open, 0 means closed).
    """
    master_valve, wing_valve, choke_valve = auto_align([master_valve, wing_valve, choke_valve], align_timesteps)

    valves = [master_valve, wing_valve, choke_valve]
    thresholds = [threshold_master, threshold_wing, threshold_choke]

    if any([i < 0 for i in thresholds]):
        raise UserValueError(
            "Threshold value has to be greater than or equal to 0"
        )  # it is not physical to have an opening less than 0

    if any([i > 100 for i in thresholds]):
        raise UserValueError(
            "Threshold value has to be less than or equal to 100"
        )  # it is not physical to have an opening more than 100

    thresholds = [  # make modifications for threshold if range is between 0-1.
        _pct_to_fraction(threshold)
        if max(valve) <= 1.0
        else float(threshold)  # if the valve series is between 0-1 then it should have a
        for threshold, valve in zip(thresholds, valves)  # max value of 1 so we change the threshold from percentage
    ]  # to fraction.
    threshold_master, threshold_wing, threshold_choke = thresholds

    is_master_gt_threshold = master_valve >= threshold_master
    is_wing_gt_threshold = wing_valve >= threshold_wing
    is_choke_gt_threshold = choke_valve >= threshold_choke

    return pd.Series(
        np.logical_and.reduce((is_master_gt_threshold, is_wing_gt_threshold, is_choke_gt_threshold)).astype(int),
        index=master_valve.index,
    )



# Copyright 2022 Cognite AS
from typing import Optional

import pandas as pd

from indsl.filter import trend_extraction_hilbert_transform
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index


@check_types
def uncertainty_rstd(
    data: pd.Series,
    resample_rate: pd.Timedelta = pd.Timedelta("30m"),
    emd_sift_thresh: float = 1e-8,
    emd_max_num_imfs: Optional[int] = None,
    emd_error_tolerance: float = 0.05,
) -> pd.Series:
    r"""Relative uncertainty.

    The relative uncertainty is computed as the ratio between the
    standard deviation of the signal noise and the mean of the true signal.
    The noise and true signals are estimated using the empirical model decomposition method.
    The relative uncertainty is computed on segments of the input data of size
    rst_resample_rate. In mathematical notation, this means:

    .. math::

        rstd = \\sigma(F_t - A_t)/|\\mu(A_t)|

    where :math:`F_t` is the resampled input time series, and :math:`A_t` is the resampled and
    detrended time series obtained using the empirical model decomposition method (EMD).

    Args:
        data: Time series
            Input time series
        resample_rate: Resample rate.
            Resample rate used when estimating the relative standard deviation
        emd_sift_thresh: Sifting threshold.
            Threshold to stop EMD sifting process. This threshold is based on the Cauchy convergence test and represents the
            residue between two consecutive oscillatory components (IMFs). A small threshold (close to zero) will
            result in more components extracted. Typically, a few IMFs are enough to build the main trend. Choosing a
            high threshold might not affect the outcome. Defaults to 1e-8.
        emd_max_num_imfs: Maximum number of components.
            Maximum number of EMD oscillatory components (or IMFs) to estimate the main trend. If no value (None) is defined
            the process continues until sifting threshold is reached. Defaults to None.
        emd_error_tolerance: Energy tolerance.
            Threshold for the EMD cross energy ratio validation used for choosing oscillatory components or IMFs. Defaults to
            0.05.

    Returns:
        pd.Series: Time series
    """
    validate_series_has_time_index(data)

    # Estimate the true signal and its noise using EDM
    truth = trend_extraction_hilbert_transform(data, emd_sift_thresh, emd_max_num_imfs, emd_error_tolerance)
    noise = data - truth

    noise_std = noise.resample(resample_rate, origin="start").std().reindex(truth.index, method="ffill").interpolate()
    truth_mean = truth.resample(resample_rate, origin="start").mean().reindex(truth.index, method="ffill").interpolate()

    relative_std = (noise_std / abs(truth_mean)).dropna()
    return relative_std



# Copyright 2021 Cognite AS
import pandas as pd

from indsl.versioning import register

from . import versioning_test_v1  # noqa


@register(version="2.0", deprecated=True, changelog="Added verbose parameter")
def versioning_test_op(series: pd.Series, verbose: bool) -> pd.Series:
    """New versioning test.

    This new function is used only for testing purposes.

    Args:
        series: Dummy input
        verbose: Dummy flag

    Returns:
        pandas.Series: Dummy output
    """
    return series



import pandas as pd

from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index, validate_series_is_not_empty, validate_timedelta_unit


@check_types
def pearson_correlation(
    data1: pd.Series,
    data2: pd.Series,
    time_window: pd.Timedelta = pd.Timedelta(minutes=15),
    min_periods: int = 1,
    align_timesteps: bool = False,
) -> pd.Series:
    """Pearson correlation.

    This function measures the linear correlation between two time series along a rolling window.
    Pearson’s definition of correlation: https://en.wikipedia.org/wiki/Pearson_correlation_coefficient

    Args:
        data1: Time series.
        data2: Time series.
        time_window: Time window.
            Length of the time period to compute the Pearson correlation. Defaults to 'minutes=15'.
            Time unit should be in days, hours, minutes or seconds. Accepted formats can be found `here
            (https://pandas.pydata.org/docs/reference/api/pandas.Timedelta.html)`_.
        min_periods: Minimum samples.
            Minimum number of observations required in the given time window (otherwise, the result is set to 0).
            Defaults to 1.
        align_timesteps (bool): Auto-align.
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
        UserTypeError: time_window is not of type pandas.Timedelta
    """
    # validations
    for data in [data1, data2]:
        validate_series_has_time_index(data)
        validate_series_is_not_empty(data)
    validate_timedelta_unit(time_window)

    # align timesteps
    data1, data2 = auto_align([data1, data2], align_timesteps)

    # compute
    return data1.rolling(time_window, min_periods=min_periods).corr(data2)



# Copyright 2021 Cognite AS
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import pandas as pd

from indsl.exceptions import UserValueError
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index


class DataQualityScore:
    """Data class storing the result of an event data quality analysis.

    Attributes:
        analysis_start: Analysis start time
        analysis_end: Analysis end time
        events: List of event events
            Represented as pairs of timestamps
    """

    @check_types
    def __init__(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        events: List[Tuple[pd.Timestamp, pd.Timestamp]],
    ):
        """Data quality score init function.

        Args:
            analysis_start: Analysis start time
            analysis_end: Analysis end time
            events: List of event events
                Represented as pairs of timestamps
        """
        for (event_start, event_end) in events:
            if event_start > event_end:
                raise UserValueError(
                    f"Expected start date of event to be before end date, got event_start='{event_start}' and event_end='{event_end}'"
                )
            if event_start < analysis_start or event_end > analysis_end:
                raise UserValueError(
                    f"Expected event to be in analysis window, got event='{event_start}-{event_end}' and analysis_window='{analysis_start}-{analysis_end}'"
                )

        self.analysis_start = analysis_start
        self.analysis_end = analysis_end
        self.events = events

    @property
    def degradation(self):
        """Degradation factors."""
        return [
            (event_end - event_start) / (self.analysis_end - self.analysis_start)
            for event_start, event_end in self.events
        ]

    @property
    def score(self) -> float:
        """Data quality score calculated as 1-sum(degradation)."""
        return 1.0 - sum(self.degradation)

    def __add__(self, other: DataQualityScore) -> DataQualityScore:
        """Return the union of two data quality results.

        Args:
            other: Other event data quality result

        Returns:
            DataQualityScore: The merged results

        Raises:
            UserValueError: If the two input results do not have a consequent analysis window
        """
        if self.analysis_end != other.analysis_start:
            raise UserValueError(
                f"Expected consecutive analysis periods in self and other, got self.analysis_end='{self.analysis_end}' and other.analysis_start='{other.analysis_start}'"
            )

        # Copy events to avoid side effects
        self_events = self.events.copy()
        other_events = other.events.copy()

        # Merge the last event of first score with the
        # first event of the second score if they are subsequent
        if len(self_events) > 0 and len(other_events) > 0 and self_events[-1][1] == other_events[0][0]:
            other_events[0] = (self_events.pop()[0], other_events[0][1])

        return DataQualityScore(self.analysis_start, other.analysis_end, self_events + other_events)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DataQualityScore):
            return (
                self.analysis_start == other.analysis_start
                and self.analysis_end == other.analysis_end
                and (np.asarray(self.events) == np.asarray(other.events)).all()
            )
        else:
            raise NotImplementedError(
                f"Equality comparison between type {type(self)} and {type(other)} not implemented"
            )


class DataQualityScoreAnalyser(ABC):
    """Object to calculate data quality scores.

    Attributes:
        series: time series
    """

    def __init__(self, series: pd.Series):
        """Data quality score analyser init function.

        Args:
            series: time series

        Raises:
            UserValueError: If series has no time index
        """
        validate_series_has_time_index(series)

        self.series = series

    @abstractmethod
    def compute_score(self, analysis_start: pd.Timestamp, analysis_end: pd.Timestamp) -> DataQualityScore:
        """Compute data quality result.

        Args:
            analysis_start: analyis start time
            analysis_end: analyis end time

        Returns:
            DataQualityScore: A DataQualityScore object

        Raises:
            UserValueError: If analysis_start < analysis_end
            UserValueError: If the analysis start and end timestamps are outside the range of the series index
        """
        if analysis_start > analysis_end:
            raise UserValueError(
                f"Expected analysis_start < analysis_end, got analysis_start '{analysis_start}' and analysis_end '{analysis_end}'"
            )

        if len(self.series) < 1:
            return  # type: ignore

        if analysis_start < self.series.index[0]:
            raise UserValueError(
                f"Expected analysis_start to be equal or after the first timestamp in series, got analysis_start={analysis_start} and series.index[0]={self.series.index[0]}"
            )
        if analysis_end > self.series.index[-1]:
            raise UserValueError(
                f"Expected analysis_end to be before or equal the last timestamp in series, got analysis_end={analysis_end} and series.index[-1]={self.series.index[-1]}"
            )
        return  # type: ignore

    @staticmethod
    def _convert_series_to_events(series) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:
        # Each gap in the input series is represented as a consecutive (1, 1) pair.
        # Hence filtering the 1 values and re-arranging the associated index as pairs
        # yields a list of the (start, end) gap events.
        events_array = list(series[series == 1].index.values.reshape(-1, 2))
        return [(pd.Timestamp(start), pd.Timestamp(end)) for start, end in events_array]

    @staticmethod
    def _filter_events_outside_analysis_period(
        events: List[Tuple[pd.Timestamp, pd.Timestamp]], analysis_start: pd.Timestamp, analysis_end: pd.Timestamp
    ) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:

        return [(start, end) for start, end in events if end > analysis_start and start < analysis_end]

    @staticmethod
    def _limit_first_and_last_events_to_analysis_period(
        gaps: List[Tuple[pd.Timestamp, pd.Timestamp]], analysis_start: pd.Timestamp, analysis_end: pd.Timestamp
    ) -> List[Tuple[pd.Timestamp, pd.Timestamp]]:

        if len(gaps) == 0:
            return gaps

        first_gap = gaps[0]
        gaps[0] = (max(first_gap[0], analysis_start), first_gap[1])

        last_gap = gaps[-1]
        gaps[-1] = (last_gap[0], min(last_gap[1], analysis_end))

        return gaps



# Copyright 2021 Cognite AS
import pandas as pd

from indsl.versioning import register


@register(version="1.0", deprecated=True)
def versioning_test_op(series: pd.Series) -> pd.Series:
    """Old versioning test.

    This old function is used only for testing purposes.

    Args:
        series: Dummy input

    Returns:
        pandas.Series: Dummy output
    """
    return series



from .accuracy.uncertainty import UncertaintyDataQualityScoreAnalyser
from .base import DataQualityScore
from .completeness.density import DensityDataQualityScoreAnalyser
from .completeness.gap import GapDataQualityScoreAnalyser



from datetime import timedelta

import numpy as np
import pandas as pd

from ..exceptions import UserValueError
from ..type_check import check_types


@check_types
def calculate_shutin_variable(
    variable_signal: pd.Series, shutin_signal: pd.Series, hrs_after_shutin: float
) -> pd.Series:
    """Shut-in variable calculator.

    The shut-in variable calculator is a function to compute the variable of interest at specific times after the shut-in
    onset. Typically, variables of interest are pressure and temperature. The function is the dependency of the shut-in
    detector. Based on the detected shut-in interval, the function uses specified number of hours indicating the time after
    the onset of each shut-in. It calculates the variable of interest at that time instance using interpolation (method - time).

    Args:
        variable_signal: Signal of interest.
            Typically pressure or temperature signal.
        shutin_signal: Shut-in signal.
            The signal comes from a shut-in detector function or a signal indicating shut-in condition
            (0 - well in shut-in state, 1 - well in flowing state). We suggest using the
            :meth:`indsl.oil_and_gas.calculate_shutin_interval`.
        hrs_after_shutin: Hours after.
            Hours after shut-in onset at which to calculate the signal of interest.

    Returns:
        pandas.Series: Output.
        Signal of interest at specific time after shut-in onset.
    """
    if shutin_signal.empty or variable_signal.empty:
        return pd.Series(dtype=np.float64, index=pd.DatetimeIndex([]))

    if shutin_signal.dtype not in [int, np.int64, np.int32]:
        raise UserValueError("The results from shut-in detector contain non-integer numbers")

    # condition1: guarantee the timeseries always start from 1 (flowing state)
    if shutin_signal.iloc[0] == 0:
        time_idx = shutin_signal[shutin_signal.values == 1].index
        if time_idx.empty:
            raise UserValueError("The signal does not contain flowing data")
        shutin_signal = shutin_signal[shutin_signal.index > time_idx[0]]

    # get changes of the shut-in signal (-1 indicates the start and 1 indicates the end of shut-in)
    changepts = shutin_signal.diff()
    ss_start_shifted = shutin_signal[changepts == -1].index + timedelta(hours=hrs_after_shutin)
    ss_end = shutin_signal[changepts == 1].index

    # condition1 guarantees that len(ss_start_shifted) is greater or equal to len(ss_end)
    if len(ss_start_shifted) > len(ss_end):
        if ss_start_shifted[-1] <= shutin_signal.index[-1]:
            ss_end = ss_end.union(shutin_signal.index[-1:])
        else:
            ss_start_shifted = ss_start_shifted[:-1]

    # condition2: filter shut-ins where shifted start is earlier than finish
    shut_ins = ss_start_shifted[ss_start_shifted <= ss_end]
    if shut_ins.empty:
        return pd.Series(dtype=np.float64, index=pd.DatetimeIndex([]))

    return (
        variable_signal.reindex(variable_signal.index.union(shut_ins))
        .interpolate(method="time")[shut_ins]
        .astype(np.float64)
        .rename(None)
    )



from .generator import const_value, insert_data_gaps, line, perturb_timestamp, sine_wave
from .noise import white_noise
from .polynomial import univariate_polynomial


TOOLBOX_NAME = "Signal generator"

__all__ = [
    "line",
    "perturb_timestamp",
    "insert_data_gaps",
    "sine_wave",
    "white_noise",
    "univariate_polynomial",
    "const_value",
]



# Copyright 2022 Cognite AS
from typing import Union

import numpy as np
import pandas as pd

from indsl import versioning
from indsl.resample.auto_align import auto_align
from indsl.ts_utils.ts_utils import scalar_to_pandas_series
from indsl.type_check import check_types
from indsl.validations import UserValueError


@versioning.register(version="1.0", deprecated=True)
@check_types
def flow_through_valve(
    inlet_P: Union[pd.Series, float],
    outlet_P: Union[pd.Series, float],
    valve_opening: Union[pd.Series, float],
    SG: float,
    type: str,
    min_opening: float,
    max_opening: float,
    min_Cv: float,
    max_Cv: float,
    align_timestamps: bool = False,
) -> pd.Series:
    r"""Valve volumetric flow rate.

    This calculation can be used when there is no flow meter, but the pressure difference over the valve is known.
    The calculated flow rate is only exactly applicable for ideal fluids (incompressible and with zero viscosity).
    The available valve characteristics are
    * Linear: :math:`Cv = ax + b`.
    * Equal percentage: :math:`Cv = ae^x + b`.
    The formula for the flow rate is
    .. math:: Q = Cv \sqrt{\frac{p_{in} - p_{out}}{SG}}.

    Args:
        inlet_P: Pressure at
            inlet [bar].
        outlet_P: Pressure at outlet
            [bar].
        valve_opening: Valve opening [-].
            Note that this is the proportional and not the percentage valve opening.
        SG: Specific gravity of
            fluid [-].
        type: Valve characteristic
            Valve characteristic, either "Linear" or "EQ" (equal percentage)
        min_opening: Min opening [-].
            Valve opening at minimum flow.
            Note that the flow coefficient should be expressed in imperial units.
        max_opening: Max opening [-].
            Valve opening at maximum flow.
            Note that the flow coefficient should be expressed in imperial units.
        min_Cv: Min Cv
            :math:`[(-, \frac{gpm}{psi^{0.5}})]`.
            Valve Cv at minimum flow.
            Note that the flow coefficient should be expressed in imperial units.
        max_Cv: Max Cv
            :math:`[(-, \frac{gpm}{psi^{0.5}})]`.
            Valve Cv at maximum flow.
            Note that the flow coefficient should be expressed in imperial units.
        align_timestamps: Auto-align.
            Automatically align time stamp  of input time series. Default is False.

    Raises:
        ValueError: If the valve characteristic is not recognized.

    Returns:
        pd.Series: Valve flow rate [m3/h].
    """
    if SG < 0:
        raise UserValueError("Specific gravity cannot be negative.")

    inlet_P, outlet_P, valve_opening = auto_align([inlet_P, outlet_P, valve_opening], align_timestamps)

    if type == "Linear":
        Cv = (max_Cv - min_Cv) / (max_opening - min_opening) * valve_opening + (
            min_Cv * max_opening - min_opening * max_Cv
        ) / (max_opening - min_opening)
    elif type == "EQ":
        exp_coef = (max_Cv - min_Cv) / (np.exp(max_opening) - np.exp(min_opening))
        intercept = (min_Cv * np.exp(max_opening) - np.exp(min_opening) * max_Cv) / (
            np.exp(max_opening) - np.exp(min_opening)
        )
        Cv = exp_coef * np.exp(valve_opening) + intercept
    else:
        raise UserValueError("Only 'Linear' or 'EQ' valve characteristics are supported.")

    Q = 0.865 * Cv * np.sqrt((inlet_P - outlet_P) / SG)

    return scalar_to_pandas_series(Q)



# Copyright 2021 Cognite AS
import pandas as pd

from indsl.versioning import register


@register(version="1.0", deprecated=True, changelog="Intended for CHARTS internal use only")
def no_op(series: pd.Series) -> pd.Series:
    """Identity operator.

    Return the input unmodified as output.

    Args:
        series: input

    Returns:
        pandas.Series: output
    """
    return series



# Copyright 2021 Cognite AS
from enum import Enum

import numpy as np
import pandas as pd

from statsmodels.tsa.arima.model import ARIMA

from indsl import versioning

from ..exceptions import UserRuntimeError, UserValueError
from ..type_check import check_types


class MethodType(Enum):
    """Method type."""

    ONESTEP = "onestep"
    MULTISTEP = "multistep"


@versioning.register(version="1.0", deprecated=True)
@check_types
def arma_predictor(
    data: pd.Series,
    ar_order: int = 2,
    ma_order: int = 2,
    train_fraction: float = 0.8,
    method: MethodType = MethodType.ONESTEP,
    steps: int = 1,
) -> pd.Series:
    """ARMA predictor.

    The ARMA predictor works by fitting constants to an auto regression (AR)  and a moving average (MA) equation and
    extrapolating the results.

    Args:
        data: Time series.
        ar_order: Number of AR terms.
        ma_order: Number of MA terms.
        train_fraction: Fraction.
            Fraction of the input data used for training the model.
        method: Method.
            Type of prediction to perform. MULTISTEP involves forecasting
            several steps ahead of the training dataset, while ONESTEP involves incrementally going through the test
            dataset, and appending it to the training dataset by performing a one-step forecast.
        steps: Steps.
            Number of steps to forecast ahead of the training dataset.

    Returns:
        pandas.Series: Prediction
        Predicted data for the test fraction of the input data (e.g., 1 - train_fraction)

    Raises:
        UserRuntimeError: If an empty time series is passed into the function.
        UserValueError: Incorrect values passed into keyword arguments. ar_order and ma_order should be integers, and train_fraction must be float between 0 and 1.
    """
    # Check length of data
    if len(data) < 1:
        raise UserRuntimeError("No data passed to algorithm.")

    # Check train-test split
    if not 0 < train_fraction < 1:
        raise UserValueError(f"train_fraction needs to be a float between 0 and 1. Got {train_fraction}")

    # Get key properties from data
    n_obs = len(data)
    n_train = int(n_obs * train_fraction)
    dt_step_sec = int(data.reset_index()["index"].diff().mean().total_seconds())

    # If method is multistep, then simply forecast some steps ahead of training dataset
    if method == MethodType.MULTISTEP:
        train_data = data.iloc[:n_train]
        pred_data = _train_and_return_forecast(ar_order, ma_order, train_data, steps, dt_step_sec)

    # If method is onestep, then iterate through test dataset and perform a one step prediction
    elif method == MethodType.ONESTEP:

        # Storage array
        store_forecast = []

        # Limiting number of cycles to 100
        cycles = min(n_obs - n_train, 100)

        for it in range(cycles + 1):

            # Append new observation
            train_data = data.iloc[it : it + n_train]
            res = _train_and_return_forecast(ar_order, ma_order, train_data, steps, dt_step_sec)
            store_forecast.append(res.tail(1))

        # Concatenate prediction dataframe
        pred_data = pd.concat(store_forecast)

    # If neither method, then raise value error
    else:
        raise UserValueError(f"Method needs to be either 'MethodType.ONESTEP' or 'MethodType.MULTISTEP'. Got {method}")

    # Return the prediction dataframe
    return pred_data.dropna()


@check_types
def _train_and_return_forecast(
    ar_order: int, ma_order: int, train_data: pd.Series, steps: int, dt_step_sec: int
) -> pd.Series:
    """Train and return forecast.

    Function to train an ARMA model and return the forecasted value as a Pandas series with a datetime index

    Args:
        ar_order: Number of AR terms in the equation
        ma_order: Number of MA terms in the equation
        train_data: data to train model on
        steps: number of steps to forecast ahead of training data
        dt_step_sec: expected granularity for each step (seconds)

    Returns:
        pandas.Series: Forecasted time series of length `steps`
    """
    # Fit ARMA model
    model = ARIMA(train_data, order=(ar_order, 0, ma_order))
    fit_model = model.fit()

    # Perform prediction and return series with datetime index
    y_pred = fit_model.forecast(steps=steps).values
    max_dt = train_data.index[-1]
    # dt_index = pd.date_range(start=max_dt, freq=str(dt_step_sec) + "s", periods=steps + 1)
    dt_index = max_dt + pd.TimedeltaIndex(np.linspace(1, steps, num=steps) * dt_step_sec, unit="s")

    return pd.Series(y_pred, index=dt_index)



# Copyright 2022 Cognite AS
import numpy as np
import pandas as pd

from indsl.ts_utils.utility_functions import (
    generate_step_series,
    modified_z_scores_test,
    normality_assumption_test,
    z_scores_test,
)

from ..type_check import check_types
from ..validations import validate_series_has_time_index, validate_series_is_not_empty


# todo add as indsl function
@check_types
def get_densities(data: pd.Series, time_window: pd.Timedelta = pd.Timedelta("5m")) -> pd.Series:
    """Point density.

    Counts the number of points on a rolling time window.

    Args:
        data: Time series.
        time_window: Window
            Length of the time period to compute the density of points. Defaults to 5min.

    Returns:
        pd.Series: Time series with the point density for each rolling window.
    """
    return data.rolling(time_window, min_periods=1).count()


@check_types
def low_density_identification_z_scores(
    data: pd.Series,
    time_window: pd.Timedelta = pd.Timedelta("5m"),
    cutoff: float = -3.0,
    test_normality_assumption: bool = False,
) -> pd.Series:
    """Low density, Z-scores.

    Detect periods with low density of data points using `Z-scores <https://en.wikipedia.org/wiki/Standard_score>`_. Z-score stands for
    the number of standard deviations by which the value of a raw score (i.e., an observed value or data point) is
    above or below the mean value of what is being observed or measured. This method assumes that the densities over a rolling window
    are normally distributed. Low density periods are defined as time periods where the Z-score is lower than cutoff.

    Args:
        data: Time series
        time_window: Rolling window.
            Length of the time period to compute the density of points. Defaults to 5 min.
        cutoff: Cut-off.
            Number of standard deviations from the mean.
            Low density periods are detected if the Z-score is below this cut-off value. Default -3.0.
        test_normality_assumption: Test for normality.
            Raises an error if the data densities over the rolling windows are not normally distributed.
            The Shapiro-Wilk test is used. The test is only performed if the time series contains less than 5000 data points.
            Default to False.

    Returns:
        pd.Series: Time series
            The returned time series is an indicator function that is 1 where there is a low density period, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: cutoff is not a number
        UserValueError: data is empty
        UserValueError: time series is not normally distributed
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    x_freq = get_densities(data, time_window=time_window)

    if test_normality_assumption:
        normality_assumption_test(series=x_freq, max_data_points=5000, min_p_value=0.05, min_W=0.5)

    is_low_density = z_scores_test(x_freq.values, cutoff=cutoff, direction="less")

    return generate_step_series(pd.Series(is_low_density, index=data.index))


@check_types
def low_density_identification_modified_z_scores(
    data: pd.Series, time_window: pd.Timedelta = pd.Timedelta("5m"), cutoff: float = -3.5
) -> pd.Series:
    """Low density, mod.Z-scores.

    Detect periods with a low density of data points using modified Z-scores.
    Low density periods are defined as time periods where the Z-score is lower than the cutoff.

    Args:
        data: Time series
        time_window: Rolling window.
            Length of the time period to compute the density of points. Defaults to 5 min.
        cutoff: Cut-off.
            Low density periods are detected if the modified Z-score is below this cut-off value. Default -3.5.

    Returns:
        pd.Series: Time series
            The returned time series is an indicator function that is 1 where there is a low density period, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: cutoff has to be of type float
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    x_freq = get_densities(data, time_window=time_window)

    is_low_density = modified_z_scores_test(x_freq.values, cutoff=cutoff, direction="less")

    return generate_step_series(pd.Series(is_low_density, index=data.index))


@check_types
def low_density_identification_iqr(data: pd.Series, time_window: pd.Timedelta = pd.Timedelta("5m")) -> pd.Series:
    """Low density, IQR.

    Detect periods with a low density of data points using the `interquartile range (IQR)
    <https://en.wikipedia.org/wiki/Interquartile_range>`_ method. The IQR is a measure of statistical
    dispersion, which is the spread of the data. Densities that are more than 1.5 IQR below Q1 are considered
    as low density periods in the data.

    Args:
        data: time series
        time_window: Rolling window.
            Length of the time period to compute the density of points. Defaults to 5 min.

    Returns:
        pd.Series: time series
            The returned time series is an indicator function that is 1 where there is a low density period, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    x_freq = get_densities(data, time_window=time_window)

    percentile25 = np.quantile(x_freq, 0.25)
    percentile75 = np.quantile(x_freq, 0.75)

    iqr = percentile75 - percentile25
    lower_limit = percentile75 - 1.5 * iqr
    is_low_density = np.where(x_freq < lower_limit, 1, 0)

    return generate_step_series(pd.Series(is_low_density, index=data.index))


@check_types
def low_density_identification_threshold(
    data: pd.Series, time_window: pd.Timedelta = pd.Timedelta("5m"), cutoff: int = 10
) -> pd.Series:
    """Low density, threshold.

    Detect periods with a low density of points using a time delta threshold as a cut-off value.

    Args:
        data: time series
        time_window: Rolling window.
            Length of the time period to compute the density of points. Defaults to 5 min.
        cutoff: Density cut-off.
            Low density periods are detected if the number of points is less than this cut-off value. Default is 10.

    Returns:
        pd.Series: time series
            The returned time series is an indicator function that is 1 where there is a low density period, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    x_freq = get_densities(data, time_window=time_window)
    check_freq_series = x_freq < cutoff
    check_freq_series = check_freq_series.astype(int)

    return generate_step_series(check_freq_series)



# Copyright 2022 Cognite AS
from typing import Tuple

import numpy as np
import pandas as pd

from indsl.exceptions import UserValueError
from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types
from indsl.validations import validate_series_is_not_empty


CONVERT_R_TO_K = 5 / 9  # conversion from R to K
CONVERT_PSI_TO_KPA = 6.8947  # conversion from psi to kpa
CONVERT_KGM3_TO_PCF = 0.062  # convion from kg/m3 to pcf
R_U = 8.314  # universal gas constant KJ/Kmol.K
M_AIR = 28.97  # Molar mass of air (Kg/kmol)


@check_types
def calculate_gas_density(
    pressure: pd.Series, temperature: pd.Series, sg: pd.Series, align_timesteps: bool = False
) -> pd.Series:
    """Gas density calculator.

    The gas density is calculated from real gas laws.The psuedo critical tempreature and pressure is
    calculated from specific gravity following `Sutton (1985) <https://doi.org/10.2118/14265-MS>`_. The
    `Beggs and Brill (1973) <https://onepetro.org/JPT/article-abstract/25/05/607/165212/A-Study-of-Two-Phase-Flow-in-Inclined-Pipes>`_
    method (explicit) is used to calculate the compressibility factor. All equations used here can be found in one place at
    `Kareem et. al. <https://link.springer.com/article/10.1007/s13202-015-0209-3>`_. The gas equation *Pv = zRT*
    is used to calculate the gas density.

    Args:
       pressure: Pressure [psi].
           Pressure time series in psi units.
       temperature: Temperature [degF].
           Temperature time series in degrees Fahrenheit units.
       sg: Specific gravity [-].
           Specific gravity of the gas.
       align_timesteps: Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
       pandas.Series: Gas density [lbm/ft3]
           Estimated gas density in pound-mass per cubic foot (pcf).

    Raises:
       UserValueError: When all values of the pressure and temperature are out of range for calculating compressibility.
    """
    # check specific gravity value
    validate_series_is_not_empty(sg)
    sg = sg[0]
    if sg == 0:
        raise UserValueError("Specific gravity cannot be zero")

    # auto-align
    pressure, temperature = auto_align([pressure, temperature], align_timesteps)

    M = M_AIR * sg  # Molar mass of gas
    R = R_U / M  # gas constant

    # calculate the pseudo - crtitical pressure and temperature
    (Pc, Tc) = calculate_critical_prop(sg)

    # convert pressure to kPa and temperature to K
    pressure_Kpa = CONVERT_PSI_TO_KPA * pressure
    temperature_K = 273 + ((temperature - 32) * (5 / 9))

    # calculate the pseudo - reduced pressure and temperature
    (Ppr, Tpr) = calculate_reduced_prop(pressure_Kpa, temperature_K, Pc, Tc)

    # calculate compressibility
    compressibility = calculate_compressibility(Ppr, Tpr)

    # check division by zero
    compressibility = compressibility[compressibility.values > 0]
    temperature_K = temperature_K[temperature_K.values > 0]
    ixs = compressibility.index.intersection(temperature_K.index)
    compressibility, temperature_K, pressure_Kpa = (
        compressibility.loc[ixs],
        temperature_K.loc[ixs],
        pressure_Kpa.loc[ixs],
    )
    if compressibility.empty or temperature_K.empty:
        raise UserValueError("Compressibility or temperature series are empty after filtering zeros")
    if R == 0:
        raise UserValueError("Gas constant is zero. Division by zero is not allowed")

    # calculate gas density
    rho = pressure_Kpa / R / temperature_K / compressibility
    rho_out = CONVERT_KGM3_TO_PCF * rho
    return rho_out.dropna()


@check_types
def calculate_critical_prop(sg: float) -> Tuple[float, float]:
    """Critical properties.

    Returns the crtitcal pressure and temperature as a function of specific gravity (Sutton - 1973).

    Args:
       sg: Specific gravity
           Specific gravity of the gas.

    Returns:
       float: Critical pressure
           Critical pressure in KPa.
       float: Critical temperature
           Critical temperature in Kelvin.
    """
    Tc_R = 169.2 + 349.5 * sg - 74 * sg**2
    Tc = CONVERT_R_TO_K * Tc_R  # convert from Rankine to K
    Pc_psi = 756.8 - 131.07 * sg - 3.6 * sg**2
    Pc = Pc_psi * CONVERT_PSI_TO_KPA  # convert from psi to Kpa
    return (Pc, Tc)


@check_types
def calculate_reduced_prop(
    pressure_Kpa: pd.Series, temperature_K: pd.Series, Pc: float, Tc: float
) -> Tuple[pd.Series, pd.Series]:
    """Reduced properties.

    Returns the reduced pressure and temperature from input pressure, temperature, and critical prperties.

    Args:
       pressure_Kpa: Pressure
           Time series containing pressure data in KPa.
       temperature_K: Temperature
           Time series containing temperature data in Kelvin.
       Pc: Critical pressure
           Critical pressure in KPa.
       Tc: Critical temperature
           Critical temperature in Kelvin.

    Returns:
       pandas.Series: Reduced pressure
           Non-dimensional series.
       pandas.Series: Reduced temperature
           Non-dimensional series.
    """
    # calculate the pseudo-reduced pressure and temperature
    Ppr = pressure_Kpa / Pc
    Tpr = temperature_K / Tc
    return (Ppr, Tpr)


@check_types
def calculate_compressibility(Ppr: pd.Series, Tpr: pd.Series) -> pd.Series:
    """Gas Compressibility.

    Returns the compressibility factor from reduced pressure and temperature.
    Also deletes points in input temperature and pressure that are out of range.

    Args:
       Ppr: Reduced pressure
            Non-dimensional series.
       Tpr: Reduced temperature
            Non-dimensional series.

    Returns:
       pandas.Series: Compressibility factor
            Non-dimensional series.

    Raises:
       UserValueError: When all values of the pressure and temperature are out of range for calculating compressibility.
    """
    # delete points that are out of range, raise error if empty
    Tpr = Tpr[(Tpr.values >= 1) & (Tpr.values <= 3)]
    Ppr = Ppr[(Ppr.values >= 0) & (Ppr.values <= 15)]
    idx = Tpr.index.intersection(Ppr.index)
    Tpr, Ppr = Tpr.loc[idx], Ppr.loc[idx]

    if Tpr.empty or Ppr.empty:
        raise UserValueError("Pressure and Temperature data are empty or out of range")

    A = 1.39 * (Tpr - 0.92) ** 0.5 - 0.36 * Tpr - 0.10
    E = 9 * (Tpr - 1)
    B = (0.62 - 0.23 * Tpr) * Ppr + ((0.066 / (Tpr - 0.86)) - 0.037) * Ppr**2 + ((0.32 * Ppr**2) / (10**E))
    C = 0.132 - 0.32 * np.log10(Tpr)
    F = 0.3106 - 0.49 * Tpr + 0.1824 * Tpr**2
    D = 10**F
    compressibility = A + ((1 - A) / np.exp(B)) + C * Ppr**D

    return compressibility



# Copyright 2021 Cognite AS
import numpy as np

from indsl.resample.auto_align import auto_align

# Trigonometric functions (using lambda because we can't inspect numpy's C funcs)
from indsl.type_check import check_types


def sin(x):
    """Sin.

    Calculates the trigonometric sine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.sin(x)


def cos(x):
    """Cos.

    Calculates the trigonometric cosine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.cos(x)


def tan(x):
    """Tan.

    Calculates the trigonometric tangent of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.tan(x)


def arcsin(x):
    """Arcsin.

    Calculates the trigonometric arcsine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arcsin(x)


def arccos(x):
    """Arccos.

    Calculates the trigonometric arccosine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arccos(x)


def arctan(x):
    """Arctan.

    Calculate inverse hyperbolic tangent of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arctan(x)


@check_types
def arctan2(x1, x2, align_timesteps: bool = False):
    """Arctan(x1, x2).

    Element-wise arc tangent of x1/x2 choosing the quadrant
    correctly.

    Args:
        x1: First time series or number
        x2: Second time series or number
        align_timesteps (bool) : Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: time series
    """
    x1, x2 = auto_align([x1, x2], align_timesteps)
    return np.arctan2(x1, x2)


def deg2rad(x):
    """Degrees to radians.

    Converts angles from degrees to radians.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.deg2rad(x)


def rad2deg(x):
    """Radians to degrees.

    Converts angles from radiants to degrees.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.rad2deg(x)


# Hyperbolic functions


def sinh(x):
    """Sinh.

    Calculates the hyperbolic sine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.sinh(x)


def cosh(x):
    """Cosh.

    Calculates the hyperbolic cosine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.cosh(x)


def tanh(x):
    """Tanh.

    Calculates the hyperbolic tangent of time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.tanh(x)


def arcsinh(x):
    """Arcsinh.

    Calculates the hyperbolic arcsine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arcsinh(x)


def arccosh(x):
    """Arccosh.

    Calculates the hyperbolic arccosine of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arccosh(x)


def arctanh(x):
    """Arctanh.

    Calculates the hyperbolic arctangent of a time series.

    Args:
        x: time series

    Returns:
        pandas.Series: time series
    """
    return np.arctanh(x)



# Copyright 2022 Cognite AS
import operator as op

import numpy as np
import pandas as pd

from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types


@check_types
def productivity_index(p_res: pd.Series, p_bh: pd.Series, Q_gas: pd.Series, align_timesteps: bool = False) -> pd.Series:
    """Productivity Index.

    The productivity index or PI is defined as the gas flow rate at the well divided by the difference in pressure
    between the reservoir and bottom hole. If no data is available for any of the inputs for a specific
    timestamp, then it will be ignored.

    Args:
        p_res: Reservoir pressure.
        p_bh: Bottomhole pressure.
        Q_gas: Gas flowrate.
        align_timesteps: Auto-align.
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Productivity index.

    Raises:
        RuntimeError: If any input time series has no data, then PI can't be computed for any timestamps.
    """
    p_res, p_bh, Q_gas = auto_align([p_res, p_bh, Q_gas], align_timesteps)

    if len(p_res) == 0 or len(p_bh) == 0 or len(Q_gas) == 0:
        raise RuntimeError("One of the inputs has no data. Please check all time series inputs.")

    prod_index = op.truediv(Q_gas, p_res - p_bh).replace([np.inf, -np.inf], np.nan).dropna()

    return prod_index.dropna()



from typing import Optional

import emd
import numpy as np
import pandas as pd

from indsl.ts_utils import get_timestamps

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_minimum_length


MIN_DATA_PT = 10


@check_types
def trend_extraction_hilbert_transform(
    series: pd.Series,
    sift_thresh: float = 1e-8,
    max_num_imfs: Optional[int] = None,
    error_tolerance: float = 0.05,
    return_trend: bool = True,
) -> pd.Series:
    """Trend / De-trend signal.

    This robust method determines the trend of any non-linear and non-stationary time series based on the `Hilbert-Huang
    Transform and empirical mode decomposition (EMD). This is a mathematically complex method and not easy to document
    in detail. See the following list of resources for more information:

        * `EMD wiki <https://en.wikipedia.org/wiki/Hilbert%E2%80%93Huang_transform>`_
        * `Trend extraction based on the Hilbert-Huang transform and Empirical Mode Decomposition
          <https://www.researchgate.net/publication/261234992_Trend_extraction_based_on_Hilbert-Huang_transform>`_
        * `EMD Python package <https://emd.readthedocs.io/en/stable/>`_ with useful examples
          and tutorials
        * `Article on EMD implementation <https://www.pnas.org/doi/full/10.1073/pnas.0701020104>`_
          applied to global surface temperature anomalies data set

    The EMD method decomposes a time series into a finite number of oscillatory components, each with a
    well-defined frequency and amplitude. These are called intrinsic mode functions (IMFs). The process of identifying
    IMFs is called sifting (i.e., filtering). The sift works by iteratively extracting oscillatory components from a
    signal. Starting from the fastest and through to the very slowest until the average envelope of the components
    is less than the sifting threshold.

    The number of components selected for building the main trend are selected using the cross energy ration
    between IMFs using the Hilbert-Huang transform to estimate the spectra. If the ratio is below a given energy
    tolerance threshold, the process stops, and the selected IMFs are added together. That is the resulting main trend.

    As an output, it is possible to select either the trend of the main signal or the de-trended signal.

    Args:
        series: Time series
        sift_thresh: Sifting threshold.
            Threshold to stop the sifting process. This threshold is based on the Cauchy convergence test and represents the
            residue between two consecutive oscillatory components (IMFs). A small threshold (close to zero) will
            result in more components being extracted. Typically, a few IMFs are enough to build the main trend. Choosing a
            high threshold may not affect the outcome. Defaults to 1e-8.
        max_num_imfs: Maximum number of components.
            Maximum number of oscillatory components (or IMFs) to estimate the main trend. If no value (None) is defined,
            the process continues until the sifting threshold is reached. Defaults to None.
        error_tolerance: Energy tolerance.
            Threshold for cross energy ratio validation used for choosing oscillatory components or IMFs. Defaults to
            0.05.
        return_trend: Output trend.
            Output the trend if true. Remove the trend from the time series if False. Defaults to True.

    Returns:
        pd.Series: Time series
    """
    validate_series_has_minimum_length(series, MIN_DATA_PT)

    if error_tolerance <= 0:
        raise UserValueError("The energy tolerance  must be higher than zero")
    if max_num_imfs is not None and max_num_imfs <= 0:
        raise UserValueError("The maximum number of oscillatory components must be an integer higher than zero")
    if sift_thresh <= 0:
        raise UserValueError("The Sifting threshold must be a number higher than zero")

    # compute IMFs
    imf = emd.sift.sift(series.values, sift_thresh, max_num_imfs)
    total_number_of_imf: int = imf.shape[1]
    index_of_the_last_imf: int = total_number_of_imf - 1

    # find sampling rate in Hz
    timestamps = get_timestamps(series, "s")
    sample_rate_hz = 1 / (np.mean(np.diff(timestamps.values)))

    # compute phase, frequency and amplitude
    phase, frequency, amplitude = emd.spectra.frequency_transform(imf, sample_rate=sample_rate_hz, method="hilbert")

    # compute Hilbert-Huang spectrum for each IMF separately and sum over time to get Hilbert marginal spectrum
    ff, hht = emd.spectra.hilberthuang(frequency, amplitude, sum_imfs=False, sample_rate=sample_rate_hz)

    # compute cross energy ratio
    significant_imf_index: int = index_of_the_last_imf

    for i in range(index_of_the_last_imf - 1, -1, -1):
        rho = np.sum(hht[:, i] * hht[:, i + 1]) / (np.sum(hht[:, i]) * np.sum(hht[:, i + 1]))

        if rho < error_tolerance:
            break
        else:
            significant_imf_index = significant_imf_index - 1

    trend = sum(imf[:, i] for i in range(significant_imf_index, total_number_of_imf))

    trend_series = pd.Series(trend, index=series.index)

    return trend_series if return_trend else series - trend_series



# Copyright 2022 Cognite AS
import pandas as pd

from indsl.ts_utils.utility_functions import generate_step_series
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index, validate_series_is_not_empty


@check_types
def unchanged_signal_detector(
    data: pd.Series, duration: pd.Timedelta = pd.Timedelta(minutes=60), min_nr_data_points: int = 3
) -> pd.Series:
    """Unchanged signal detection.

    Detect periods of time when the data stays at a constant value for longer than a given time window.

    Args:
        data: Time series.
        duration: Time window.
            Length of the time period to check for unchanged time series values. Defaults to 'minutes=60'.
        min_nr_data_points: Data points.
            The least number of data points to avoid alerting on missing data. Defaults to 3.

    Returns:
        pd.Series: Time series.
            The returned time series is an indicator function that is 1 where the time series value has remained
            unchanged, 0 if it has changed.

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    # Convert series to dataframe
    df = data.to_frame()

    duration_in_hours = duration.total_seconds() / 3600

    # Rename dataframe columns
    df.columns = ["value"]
    df["datetime"] = df.index

    # Calculate cumulative sum of consecutive time series value differences to see how much the values change over time
    df["value_grp"] = (df.value.diff(1) != 0).astype("int").cumsum()

    # Count the number of consecutive unchanged values and their start and end time stamps
    df1 = pd.DataFrame(
        {
            "BeginDate": df.groupby("value_grp").datetime.first(),
            "EndDate": df.groupby("value_grp").datetime.last(),
            "Consecutive": df.groupby("value_grp").size(),
        }
    ).reset_index(drop=True)

    # Calculate the time between unchanged signal values - in hours
    df1["time_passed"] = (df1["EndDate"] - df1["BeginDate"]).dt.total_seconds() / 3600

    # Merge the original dataframe with the new based on 'datetime' and 'BeginDate'
    df = pd.merge(df, df1, how="left", left_on="datetime", right_on="BeginDate")

    df["is_unchanged"] = 0

    # If the time passed between the occurrence of unchanged data points in the signal is greater than the provided
    # duration, set 'is_unchanged' to 1, else 0.
    # Also check that there are at least 'min_nr_data_points' data points to avoid alerting about missing data
    # The default number of data points is set to 3 so that we have at least one data point between start and stop

    for i in range(len(df)):
        if df["time_passed"].iloc[i] >= duration_in_hours and df["Consecutive"].iloc[i] >= min_nr_data_points:
            df.at[i, "is_unchanged"] = 1

    # Set value of 'is_unchanged' to 1 for the whole duration of unchanged signal
    for i in range(len(df)):
        if df["is_unchanged"][i] == 1:
            j = 1
            while j < df["Consecutive"].iloc[i] and i + j < len(df):
                df.at[i + j, "is_unchanged"] = 1
                j = j + 1

    df = df[["datetime", "is_unchanged"]]

    # Convert dataframe to series
    df = df.set_index("datetime")
    data = df.squeeze()

    return generate_step_series(data)



from functools import wraps

from typeguard import typechecked

from indsl.exceptions import UserTypeError


def error_handling(operation):
    """Decorator that catches TypeError and wraps to inDSL specific error."""

    @wraps(operation)
    def wrapper(*args, **kwargs):
        try:
            return operation(*args, **kwargs)
        except TypeError as e:
            raise UserTypeError(str(e)) from e

    return wrapper


def check_types(operation):
    """Decorator to check types of inputs and outputs of a function.

    Decorator uses typeguard library to validate arguments of a
    function, and then wraps a TypeError to UserTypeError which is
    specific for inDSL library
    """
    return error_handling(typechecked(operation))



# Copyright 2022 Cognite AS

from datetime import datetime, timedelta
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd

from scipy.stats import shapiro

from indsl.resample.auto_align import auto_align

from ..exceptions import UserValueError
from ..type_check import check_types


TimeUnits = Literal["ns", "us", "ms", "s", "m", "h", "D", "W"]


# Rounding and utility functions
@check_types
def round(x, decimals: int):
    """Round.

    Rounds a time series to a given number of decimals.

    Args:
        x: time series
        decimals: number of decimals

    Returns:
        pd.Series: time series

    """
    return np.round_(x, decimals=decimals)


def floor(x):
    """Round down.

    Rounds a time series down to the nearest integer smaller than
    or equal to the current value.

    Args:
        x: time series

    Returns:
        pd.Series: time series
    """
    return np.floor(x)


def ceil(x):
    """Round up.

    Rounds up a time series to the nearest integer greater than or
    equal to the current value.

    Args:
        x: time series

    Returns:
        pd.Series: time series
    """
    return np.ceil(x)


def sign(x):
    """Sign.

    Element-wise indication of the sign of a time series.

    Args:
        x: time series

    Returns:
        pd.Series: time series
    """
    return np.sign(x)


# Clipping functions
@check_types
def clip(x, low: float = -np.inf, high: float = np.inf):
    """Clip (low, high).

    Given an interval, values of the time series outside the
    interval are clipped to the interval edges.

    Args:
        x (pd.Series): time series
        low: Lower limit
            Lower clipping limit. Default: -infinity
        high: Upper limit
            Upper clipping limit. Default: +infinity

    Returns:
        pd.Series: time series
    """
    return np.clip(x, low, high)


@check_types
def maximum(x1, x2, align_timesteps: bool = False):
    """Element-wise maximum.

    Computes the maximum value of two time series or
    numbers.

    Args:
        x1: First time series or number
        x2: Second time series or number
        align_timesteps: Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        pd.Series: time series
    """
    x1, x2 = auto_align([x1, x2], align_timesteps)
    return np.maximum(x1, x2)


@check_types
def minimum(x1, x2, align_timesteps: bool = False):
    """Element-wise minimum.

    Computes the minimum value of two time series.

    Args:
        x1: First time series or number
        x2: Second time series or number
        align_timesteps: Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        pd.Series: time series
    """
    x1, x2 = auto_align([x1, x2], align_timesteps)
    return np.minimum(x1, x2)


@check_types
def union(series1: pd.Series, series2: pd.Series) -> pd.Series:
    """Union.

    Takes the union of two time series. If a time stamp
    occurs in both series, the value of the first time series is used.

    Args:
        series1: time series
        series2: time series

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: series1 or series2 is not a time series
    """
    out = pd.concat([series1, series2]).sort_index()
    return out[~out.index.duplicated(keep="first")]


@check_types
def bin_map(x1, x2, align_timesteps: bool = False):
    """Element-wise greater-than.

    Maps to a binary array by checking if one time series is greater than another.

    Args:
        x1: First time series or number
        x2: Second time series or number
        align_timesteps: Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        np.ndarray: time series
    """
    x1, x2 = auto_align([x1, x2], align_timesteps)
    return out.astype(np.int64) if hasattr(out := x1 > x2, "astype") else out


@check_types
def set_timestamps(timestamp_series: pd.Series, value_series: pd.Series, unit: TimeUnits = "ms") -> pd.Series:
    """Set index of time series.

    Sets the time series values to the Unix timestamps.
    The timestamps follow the Unix convention (Number of seconds
    starting from January 1st, 1970). Both input time series
    must have the same length.


    Args:
        timestamp_series: Timestamp time series
        value_series: Value time series
        unit: Timestamp unit
          Valid values "ns|us|ms|s|m|h|D|W". Default "ms"

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: timestamp_series or value_series are not time series
        UserTypeError: unit is not a string
        UserValueError: timestamp_series and value_series do not have the same length
    """
    if not len(timestamp_series) == len(value_series):
        raise UserValueError("Length of input time series must be equal.")

    index = pd.to_datetime(timestamp_series.to_numpy(), unit=unit)
    return pd.Series(value_series.to_numpy(), index=index)


@check_types
def get_timestamps(series: pd.Series, unit: TimeUnits = "ms") -> pd.Series:
    """Get index of time series.

    Get timestamps of the time series as values.
    The timestamps follow the Unix convention (Number of seconds
    starting from January 1st, 1970). Precision loss in the order of
    nanoseconds may happen if unit is not nanoseconds.

    Args:
        series: Time-series
        unit: Timestamp unit
          Valid values "ns|us|ms|s|m|h|D|W". Default "ms"

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: series is not a time series
        UserTypeError: unit is not a string
    """
    if unit == "ns":
        values = series.index.to_numpy("datetime64[ns]").view(np.int64)
    else:
        values = series.index.view(np.int64) / pd.Timedelta(1, unit=unit).value

    return pd.Series(values, index=series.index)


@check_types
def time_shift(series: pd.Series, n_units: float = 0, unit: TimeUnits = "ms") -> pd.Series:
    """Shift time series.

    Shift time series by a time period

    Args:
        series: Time-series
        n_units: Time periods to shift
            Number of time periods to shift
        unit: Time period unit
          Valid values "ns|us|ms|s|m|h|D|W". Default "ms"

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: series is not a time series
        UserTypeError: n_units is not a number
        UserTypeError: unit is not a string
    """
    out = series.copy()
    out.index += pd.Timedelta(n_units, unit=unit)

    return out


@check_types
def replace(series: pd.Series, to_replace: Optional[List[float]] = None, value: Optional[float] = 0.0):
    """Replace.

    Replace values in a time series. The values to replace should be
    a semicolon-separated list. Undefined and infinity values can be replaced
    by using nan, inf and -inf (e.g. 1.0, 5, inf, -inf, 20, nan).

    Args:
        series: Time series
        to_replace: Replace
            List of values to replace. The values must be seperated by semicolons. Infinity and undefined values can be
            replaced by using the keywords inf, -inf and nan. The default is to replace no values.
        value: By
            Value used as replacement. Default is 0.0

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: series is not a time series
        UserTypeError: to_replace is not a list
        UserTypeError: value is not a number

    """
    if to_replace is None:
        return series
    return series.replace(to_replace, value)


@check_types
def remove(
    series: pd.Series,
    to_remove: Optional[List[float]] = None,
    range_from: Optional[float] = None,
    range_to: Optional[float] = None,
):
    """Remove.

    Remove specific values or a range of values from a time series.

    Args:
        series: Time series
        to_remove: Values
            List of values to remove. The values must be seperated by semicolons. Infinity and undefined values can be
            replaced by using the keywords inf, -inf and nan. If empty, which is the default, all values are kept.
        range_from: Range start
            Only values above this parameter will be kept. If empty, which is the default, this range filter is deactivated.
        range_to: Range end
            Only values above this parameter will be kept. If empty, which is the default, the range filter is deactivated.

    Returns:
        pd.Series: time series

    Raises:
        UserTypeError: series is not a time series
        UserTypeError: to_remove is not a list
    """
    if to_remove is not None:
        series = series[~series.isin(to_remove)]
    if range_from is not None:
        series = series[series >= range_from]
    if range_to is not None:
        series = series[series <= range_to]
    return series


@check_types
def threshold(series: pd.Series, low: float = -np.inf, high: float = np.inf):
    """Threshold.

    Indicates if the input series exceeds the lower and higher limits. The output series
    is 1.0 if the input is between the (inclusive) limits, and 0.0 otherwise.

    Args:
        series: Time series
        low: Lower limit
           threshold. Default: -infinity
        high: Upper limit
           threshold. Default: +infinity

    Returns:
        pd.Series: Time series

    Raises:
        UserTypeError: series is not a time series
        UserTypeError: low or high are not floats
    """
    return series.between(low, high).astype(np.int64)


def generate_step_series(flag: pd.Series) -> pd.Series:
    """Step-wise time series.

    Construct a step-wise time series (with 0-1 values) from a flag time series.

    Args:
        flag: Binary time series.
            The length of the flag time series has to be the same as the original data points.

    Returns:
        pd.Series: Time series

    Example:
        Given 4 datapoints and a flag with values [0, 0, 1, 0]

            data points: x-------x-------x-------x
            flag:        0       0       1       0

        The resulting step series is represented as:

                                         x-------x
            step series: x---------------x
                         0               01      1
    """
    # add first point as the start of the step time series
    step_time_series = pd.Series([flag[0]], index=[flag.index[0]])

    # compute actual_value - previous_value
    flag_shift = flag.diff(1)

    # find end points of 1-0 (decreasng) or 0-1 (increasing) patterns
    end_point_increasing = flag_shift[flag_shift == 1]
    end_point_decreasing = flag_shift[flag_shift == -1]

    # add "1" value for 0-1 (increasing) pattern, add "0" value for 1-0 (decreasing) pattern
    end_point_decreasing_values = pd.Series(
        [0] * len(end_point_decreasing), index=end_point_decreasing.index, dtype=np.int64
    )
    end_point_increasing_values = pd.Series(
        [1] * len(end_point_increasing), index=end_point_increasing.index, dtype=np.int64
    )

    # add extra "corner" points
    extra_point_decreasing_values = pd.Series(
        [1] * len(end_point_decreasing), index=end_point_decreasing.index - pd.Timedelta(1, unit="ms"), dtype=np.int64
    )
    extra_point_increasing_values = pd.Series(
        [0] * len(end_point_increasing), index=end_point_increasing.index - pd.Timedelta(1, unit="ms"), dtype=np.int64
    )

    step_time_series = pd.concat(
        [
            step_time_series,
            end_point_increasing_values,
            end_point_decreasing_values,
            extra_point_increasing_values,
            extra_point_decreasing_values,
        ]
    )
    step_time_series = step_time_series.sort_index()

    # add last point if it doesn't exist
    if flag.index[-1] != step_time_series.index[-1]:
        step_time_series = pd.concat([step_time_series, pd.Series(flag[-1], index=[flag.index[-1]])])

    # remove the last point if the number of 1 values is odd
    if len(step_time_series[step_time_series == 1]) % 2 != 0:
        step_time_series = step_time_series[:-1]

    return step_time_series


def create_series_from_timesteps(timesteps: List[timedelta]) -> pd.Series:
    """Time series from timestamps.

    Create a time series which starts on the 2021-1-1 and contains data points
    with the provided timesteps in seconds. All values of the output series are zero.

    Args:
        timesteps: Time steps between points.

    Returns:
        pd.Series: Time series
    """
    current = datetime(2021, 1, 1)
    timestamps = [current]
    for step in timesteps:
        timestamps.append(timestamps[-1] + step)
    timestamps = pd.Series(timestamps)
    return pd.Series(np.zeros(len(timestamps)), index=timestamps)


def normality_assumption_test(
    series: Union[pd.Series, np.array], max_data_points: int = 5000, min_p_value: float = 0.05, min_W: float = 0.5
) -> Tuple[float, float]:
    """Test for normality assumption.

    This function performs a Shapiro-Wilk test to check if the data is normally distributed.

    Args:
        series: Time series
        max_data_points: Maximum number of data points.
            The test is only performed if the time series contains less data points than this value.
            Default to 5000.
        min_p_value: Minimum p value.
            Probability of the time series not being normally distributed.
            Default to 0.05.
        min_W: Minimum W value.
            W is between 0 and 1, small values lead to a rejection of the
            normality assumption. Ref: https://www.nrc.gov/docs/ML1714/ML17143A100.pdf
            Default to 0.5.

    Raises:
        UserValueError: time series has more than the maximum number of datapoints allowed for this test.
        UserValueError: time series is uniform and not normally distributed
        UserValueError: time series is not normally distributed.
    """
    if len(series) > max_data_points:
        raise UserValueError(f"This time series exceeds the limit of {max_data_points} datapoints to perform the test.")

    W, p_value = shapiro(series)

    if p_value == 1.0 and W == 1.0:
        raise UserValueError("This time series is uniform and not normally distributed")
    if p_value < min_p_value or W < min_W:
        raise UserValueError("This time series is not normally distributed")

    return  # type: ignore


def z_scores_test(x: np.array, cutoff: float = 3.0, direction: Literal["greater", "less"] = "greater") -> np.array:
    """Z-scores test.

    This functions performs a z-scores test given a cut-off value and returns a binary time series.

    Args:
        x: Data values.
        cutoff: Cut-off
            Number of standard deviations from the mean.
        direction: Direction of the test.
            Options:
                "greater": the function will check if the z-scores are greater than the cut-off
                "less": the function will check if the z-scores are less than the cut-off

    Returns:
         np.array: Binary np.array with the results of the test.
            1 values indicate that the test has passed.
    """
    if (std := x.std()) == 0.0:
        z_scores = np.zeros(len(x))
    else:
        z_scores = (x - x.mean()) / std

    return threshold_test(z_scores, cutoff, direction)


def modified_z_scores_test(
    x: np.array, cutoff: float = 3.0, direction: Literal["greater", "less"] = "greater"
) -> np.array:
    """Z-scores test.

    This functions performs a modified z-scores test given a cut-off value and returns a binary time series.

    Args:
        x: Data values.
        cutoff: Cut-off value.
        direction: Direction of the test.
            Options:
                "greater": the function will check if the modified z-scores are greater than the cut-off
                "less": the function will check if the modified z-scores are less than the cut-off

    Returns:
         np.array: Binary np.array with the results of the test.
            1 values indicate that the test has passed.
    """
    median = np.median(x)
    if (mad := np.median(np.abs(x - median))) == 0.0:
        modified_z_scores = np.zeros(len(x))
    else:
        modified_z_scores = 0.6745 * (x - median) / mad

    return threshold_test(modified_z_scores, cutoff, direction)


def iqr_test(
    x: np.array, limit: Literal["upper", "lower"] = "upper", direction: Literal["greater", "less"] = "greater"
) -> np.array:
    """Z-scores test.

    This functions performs a iqr test and returns a binary time series.

    Args:
        x: Data values.
        limit: Cut-off limit.
        direction: Direction of the test.
            Options:
                "greater": the function will check if the values are greater than the cut-off limit.
                "less": the function will check if the values are less than the cut-off limit.

    Returns:
         np.array: Binary np.array with the results of the test.
            1 values indicate that the test has passed.
    """
    percentile25 = np.quantile(x, 0.25)
    percentile75 = np.quantile(x, 0.75)
    iqr = percentile75 - percentile25

    if limit == "upper":
        cutoff = percentile75 + 1.5 * iqr
    elif limit == "lower":
        cutoff = percentile25 - 1.5 * iqr

    return threshold_test(x, cutoff, direction)


def threshold_test(x: np.array, cutoff: float, direction: Literal["greater", "less"]) -> np.array:
    """Threshold test.

    This functions performs a threshold test based on a direction (greater than or less than).

    Args:
        x: Data values.
        cutoff: Cut-off value.
        direction: Direction of the test.
            Options:
                "greater": the function will check if the values are greater than the cut-off limit.
                "less": the function will check if the values are less than the cut-off limit.

    Returns:
         np.array: Binary np.array with the results of the test.
            1 values indicate that the test has passed.
    """
    if direction == "greater":
        return np.where(x > cutoff, 1, 0)
    elif direction == "less":
        return np.where(x < cutoff, 1, 0)



# Copyright 2022 Cognite AS
import pandas as pd

from scipy.signal import savgol_filter

from ..exceptions import UserValueError

# noinspection SpellCheckingInspection
from ..type_check import check_types


@check_types
def sg(data: pd.Series, window_length: int = None, polyorder: int = 1) -> pd.Series:
    """Saviztky-Golay.

    Use this filter for smoothing data without distorting the data tendency. The method is independent of
    the sampling frequency. Hence, it is simple and robust to apply to data with non-uniform sampling. If you work with
    high-frequency data (e.g., sampling frequency ~> 1 Hz), we recommend that you provide the filter window length and
    polynomial order parameters to suit the requirements. Otherwise, if no parameters are provided, the function will
    estimate and set the parameters based on the characteristics of the input time series (e.g., sampling frequency).

    Args:
        data: Time series.
        window_length: Window.
            Point-wise length of the filter window (i.e., number of data points). A large window results in a stronger
            smoothing effect and vice-versa. If the filter window_length is not defined by the user, a
            length of about 1/5 of the length of time series is set.
        polyorder: Polynomial order.
            Order of the polynomial used to fit the samples. Must be less than filter window length. Defaults to 1.
            Hint: A small polyorder (e.g., polyorder = 1) results in a stronger data smoothing effect, representing the
            dominating trend and attenuating data fluctuations.

    Returns:
        pandas.Series: Smoothed time series.

    Raises:
        UserValueError: The window length must be a positive odd integer
        UserValueError: The window length must be less than or equal to the number of data points in your time series
        UserValueError: The polynomial order must be less than the window length
    """
    if window_length is None:
        # window_length calculation
        # Filter window based on the average number of data points per second
        # avg_dps = np.mean(np.diff(series.index)) / np.timedelta64(1, "s")
        window_length = len(data) // 5
    if window_length % 2 == 0:
        window_length += 1  # The filter requires the window to be odd
    if window_length <= 0:
        raise UserValueError("The window length must be a positive odd integer.")
    if window_length > len(data):
        raise UserValueError(
            "The window length must be less than or equal to the number of data points in your time series."
        )
    if polyorder >= window_length:
        raise UserValueError("The polynomial order must be less than the window length.")
    return pd.Series(savgol_filter(data, window_length=window_length, polyorder=polyorder), index=data.index)



# Copyright 2021 Cognite AS
import pandas as pd

from indsl.type_check import check_types


@check_types
def Re(speed: pd.Series, density: float, d_viscosity: float, length_scale: float) -> pd.Series:
    """Reynolds Number.

    The Reynolds number is the ratio of inertial forces to viscous forces within a fluid subjected to
    relative internal movement due to different fluid velocities.

    Args:
        speed: Flow speed.
        density: Density.
            Density of the fluid.
        d_viscosity: Dynamic viscosity.
            Dynamic viscosity of the fluid.
        length_scale: Characteristic length.
            Characteristic linear dimension. A characteristic length is an important dimension that defines the scale
            of a physical system. Often, the characteristic length is the volume of a system divided by its surface.

    Returns:
        pandas.Series: Reynolds number
    """
    return speed * density * length_scale / d_viscosity



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['indsl',
 'indsl.data_quality',
 'indsl.data_quality.score',
 'indsl.data_quality.score.accuracy',
 'indsl.data_quality.score.completeness',
 'indsl.detect',
 'indsl.equipment',
 'indsl.filter',
 'indsl.fluid_dynamics',
 'indsl.forecast',
 'indsl.not_listed_operations',
 'indsl.oil_and_gas',
 'indsl.regression',
 'indsl.resample',
 'indsl.signals',
 'indsl.smooth',
 'indsl.statistics',
 'indsl.sustainability',
 'indsl.ts_utils']

package_data = \
{'': ['*'], 'indsl.smooth': ['img/*'], 'indsl.statistics': ['img/*']}

install_requires = \
['PyWavelets>=1.2.0,<2.0.0',
 'csaps>=1.1.0,<2.0.0',
 'emd>=0.5.3,<0.6.0',
 'flake8-docstrings>=1.6.0,<2.0.0',
 'kneed>=0.8.0,<0.9.0',
 'numba>=0.55.0,<0.56.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.3.5,<2.0.0',
 'pytest>=6.2.2,<8.0.0',
 'scikit-image>=0.19.0,<0.20.0',
 'scipy>=1.7.3,<2.0.0',
 'sklearn>=0.0,<0.1',
 'statsmodels>=0.13.1,<0.14.0',
 'toml>=0.10.2,<0.11.0',
 'typeguard>=2.13.3,<3.0.0']

setup_kwargs = {
    'name': 'indsl',
    'version': '6.2.0',
    'description': 'Industrial Data Science Library by Cognite',
    'long_description': '[![Downloads](https://static.pepy.tech/personalized-badge/indsl?period=total&units=international_system&left_color=black&right_color=brightgreen&left_text=PyPi%20Downloads)](https://pepy.tech/project/indsl) [![Code Quality](https://github.com/cognitedata/indsl/actions/workflows/code-quality.yaml/badge.svg)](https://github.com/cognitedata/indsl/actions/workflows/code-quality.yaml) [![CodeQL](https://github.com/cognitedata/indsl/actions/workflows/codeql-analysis.yml/badge.svg)](https://github.com/cognitedata/indsl/actions/workflows/codeql-analysis.yml) [![codecov](https://codecov.io/gh/cognitedata/indsl/branch/master/graph/badge.svg?token=N63jUovh1o)](https://codecov.io/gh/cognitedata/indsl)<!-- ALL-CONTRIBUTORS-BADGE:START - Do not remove or modify this section -->\n[![All Contributors](https://img.shields.io/badge/all_contributors-11-orange.svg?style=flat-square)](#contributors-)\n<!-- ALL-CONTRIBUTORS-BADGE:END -->\n\n<a href="https://cognite.com/">\n    <img src="https://github.com/cognitedata/cognite-python-docs/blob/master/img/cognite_logo.png" alt="Cognite logo" title="Cognite" align="right" height="80" />\n</a>\n\nIndustrial Data Science Library\n=========================================\n\nThis is Cognite\'s collection of data science algorithms and models. Its objective is twofold. First, empower domain\nexperts to conduct exploratory work, root cause analysis, and analyze data via <a href="https://charts.cogniteapp.com/" target="_blank">Cognite Charts</a>.\nSecond, curate a collection industry relevant data science algorithms to be used as a regular python package.\nFor more information, consult the <a href="https://docs.cognite.com/cdf/charts/" target="_blank">Charts documentation page</a>.\n\n## Documentation\n* [InDSL Home](https://indsl.docs.cognite.com/)\n* [Data Science Developer Guidelines](https://indsl.docs.cognite.com/contribute.html)\n* [Gallery of Charts - Examples](https://indsl.docs.cognite.com/auto_examples/index.html)\n* [Publishing new versions of InDSL](./PUBLISHING.md)\n\n## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://github.com/gzarruk"><img src="https://avatars.githubusercontent.com/u/24595022?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Gustavo</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=gzarruk" title="Documentation">📖</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3Agzarruk" title="Reviewed Pull Requests">👀</a> <a href="#talk-gzarruk" title="Talks">📢</a> <a href="https://github.com/cognitedata/indsl/commits?author=gzarruk" title="Tests">⚠️</a> <a href="#data-gzarruk" title="Data">🔣</a> <a href="#content-gzarruk" title="Content">🖋</a> <a href="https://github.com/cognitedata/indsl/commits?author=gzarruk" title="Code">💻</a> <a href="#ideas-gzarruk" title="Ideas, Planning, & Feedback">🤔</a></td>\n    <td align="center"><a href="https://github.com/funsim"><img src="https://avatars.githubusercontent.com/u/763150?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Simon Funke</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=funsim" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/commits?author=funsim" title="Tests">⚠️</a> <a href="https://github.com/cognitedata/indsl/issues?q=author%3Afunsim" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/neringaalt"><img src="https://avatars.githubusercontent.com/u/8692658?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Neringa Altanaite</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=neringaalt" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3Aneringaalt" title="Reviewed Pull Requests">👀</a> <a href="https://github.com/cognitedata/indsl/commits?author=neringaalt" title="Tests">⚠️</a> <a href="https://github.com/cognitedata/indsl/commits?author=neringaalt" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://www.linkedin.com/in/rhuanbarreto/"><img src="https://avatars.githubusercontent.com/u/283004?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Rhuan Barreto</b></sub></a><br /><a href="#tool-rhuanbarreto" title="Tools">🔧</a> <a href="#security-rhuanbarreto" title="Security">🛡️</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3Arhuanbarreto" title="Reviewed Pull Requests">👀</a> <a href="#ideas-rhuanbarreto" title="Ideas, Planning, & Feedback">🤔</a> <a href="#infra-rhuanbarreto" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n    <td align="center"><a href="https://github.com/evertoncolling"><img src="https://avatars.githubusercontent.com/u/33816483?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Everton Colling</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=evertoncolling" title="Code">💻</a> <a href="#data-evertoncolling" title="Data">🔣</a> <a href="https://github.com/cognitedata/indsl/commits?author=evertoncolling" title="Documentation">📖</a> <a href="https://github.com/cognitedata/indsl/commits?author=evertoncolling" title="Tests">⚠️</a></td>\n    <td align="center"><a href="https://github.com/redzarosliCognite"><img src="https://avatars.githubusercontent.com/u/91888036?v=4?s=100" width="100px;" alt=""/><br /><sub><b>redzarosliCognite</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=redzarosliCognite" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3AredzarosliCognite" title="Reviewed Pull Requests">👀</a> <a href="#ideas-redzarosliCognite" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/cognitedata/indsl/commits?author=redzarosliCognite" title="Documentation">📖</a> <a href="#example-redzarosliCognite" title="Examples">💡</a> <a href="#question-redzarosliCognite" title="Answering Questions">💬</a> <a href="#userTesting-redzarosliCognite" title="User Testing">📓</a></td>\n    <td align="center"><a href="https://github.com/MortGron"><img src="https://avatars.githubusercontent.com/u/42722577?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Morten Grønbech</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=MortGron" title="Code">💻</a> <a href="#example-MortGron" title="Examples">💡</a> <a href="https://github.com/cognitedata/indsl/commits?author=MortGron" title="Documentation">📖</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3AMortGron" title="Reviewed Pull Requests">👀</a> <a href="#question-MortGron" title="Answering Questions">💬</a> <a href="#ideas-MortGron" title="Ideas, Planning, & Feedback">🤔</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/Anitsirc22"><img src="https://avatars.githubusercontent.com/u/38993790?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Cristina Ferrer</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=Anitsirc22" title="Code">💻</a> <a href="#example-Anitsirc22" title="Examples">💡</a> <a href="#ideas-Anitsirc22" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3AAnitsirc22" title="Reviewed Pull Requests">👀</a></td>\n    <td align="center"><a href="http://treiderphoto.com"><img src="https://avatars.githubusercontent.com/u/8521241?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Håkon V. Treider</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=haakonvt" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3Ahaakonvt" title="Reviewed Pull Requests">👀</a> <a href="#ideas-haakonvt" title="Ideas, Planning, & Feedback">🤔</a> <a href="#infra-haakonvt" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n    <td align="center"><a href="https://github.com/KeepFloyding"><img src="https://avatars.githubusercontent.com/u/29730122?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Andris Piebalgs</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=KeepFloyding" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/pulls?q=is%3Apr+reviewed-by%3AKeepFloyding" title="Reviewed Pull Requests">👀</a> <a href="#example-KeepFloyding" title="Examples">💡</a></td>\n    <td align="center"><a href="https://github.com/kbrattli"><img src="https://avatars.githubusercontent.com/u/45734104?v=4?s=100" width="100px;" alt=""/><br /><sub><b>kbrattli</b></sub></a><br /><a href="https://github.com/cognitedata/indsl/commits?author=kbrattli" title="Code">💻</a> <a href="https://github.com/cognitedata/indsl/commits?author=kbrattli" title="Documentation">📖</a> <a href="#maintenance-kbrattli" title="Maintenance">🚧</a> <a href="https://github.com/cognitedata/indsl/commits?author=kbrattli" title="Tests">⚠️</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome!\n',
    'author': 'Gustavo Zarruk',
    'author_email': 'gustavo.zarruk@cognite.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://indsl.docs.cognite.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Copyright 2022 Cognite AS
import pandas as pd

from scipy.signal import butter, lfilter, sosfilt, zpk2tf

from indsl import versioning
from indsl.exceptions import UserValueError
from indsl.type_check import check_types


# noinspection SpellCheckingInspection
@check_types
@versioning.register(version="1.0", deprecated=True)
def butterworth(
    data: pd.Series, N: int = 50, Wn: float = 0.1, output: str = "sos", btype: str = "lowpass"
) -> pd.Series:
    """Butterworth.

    This signal processing filter is designed to have a
    frequency response as flat as possible in the passband and roll-offs
    towards zero in the stopband.

    In other words, this filter is designed not to modify much the signal at the
    in the passband and attenuate as much as possible the signal at the stopband. At the moment, only low and high pass
    filtering are supported.

    Args:
        data: Time series.
        N: Order.
            Defaults to 50.
        Wn: Critical frequency.
            Number between 0 and 1, with 1 representing one-half of the sampling rate (Nyquist frequency).
            Defaults to 0.1.
        output: Filtering method
            Defaults to "sos".
        btype: Filter type
            The options are: "lowpass" and "highpass"
            Defaults to "lowpass".

    Returns:
        pandas.Series: Filtered signal.
    """
    if output == "sos":
        # Create Butterworth fi
        filter_output = butter(N=N, Wn=Wn, output=output, btype=btype)
        # Apply second order segments
        filtered = sosfilt(filter_output, data, axis=0)
    elif output == "ba":
        # Create Butterworth filter
        b, a = butter(N=N, Wn=Wn, output=output, btype=btype)
        # Apply filter
        filtered = lfilter(b, a, data, axis=0)
    elif output == "zpk":
        # Create Butterworth filter
        z, p, k = butter(N=N, Wn=Wn, output=output, btype=btype)
        # Return polynomial transfer function representation from zeros and poles
        b, a = zpk2tf(z, p, k)
        # Apply filter
        filtered = lfilter(b, a, data, axis=0)
    else:
        raise UserValueError("output argument is not 'sos', 'ba' or 'zpk'.")

    return pd.Series(filtered, index=data.index)



# Copyright 2022 Cognite AS
import pandas as pd

from scipy.signal import cheby1, cheby2, sosfilt

from ..exceptions import UserValueError
from ..type_check import check_types


# noinspection SpellCheckingInspection


@check_types
def chebyshev(
    data: pd.Series,
    filter_type: int = 1,
    N: int = 10,
    rp: float = 0.1,
    Wn: float = 0.1,
    btype: str = "lowpass",
) -> pd.Series:
    """Chebyshev (I, II).

    Chebyshev filters are analog or digital filters having a steeper roll-off than Butterworth filters, and have
    passband ripple (type I) or stopband ripple (type II). Chebyshev filters have the property that they minimize the
    error between the idealized and the actual filter characteristic over the range of the filter but with ripples in
    the passband (Wikipedia).

    Args:
        data: Time series.
        filter_type: Filter type
            Options are 1 or 2. Defaults to 1.
        N: Order
            Defaults to 10.
        rp: Maximum ripple.
            Maximum ripple allowed below unity gain in the passband.
            Defaults to 0.1.
        Wn: Critical frequency.
            Number between 0 and 1, with 1 representing one-half of the sampling rate (Nyquist frequency).
            Defaults to 0.1.
        btype: Filter type.
            The options are: "lowpass" and "highpass"
            Defaults to "lowpass".

    Returns:
        pandas.Series: Filtered signal
    """
    # TODO: Move function to filter toolbox
    # TODO: Implement band and stop pass types
    # noqa
    # Only type 1 and 2 chebyshev filter exist
    if filter_type not in {1, 2}:
        raise UserValueError("Filter type must be either 1 or 2.")

    if len(data) < 1:
        return data

    # Get type 1 and 2 chebyshev filter
    cheby_filters = {1: cheby1, 2: cheby2}

    # Get filter output
    filter_output = cheby_filters[filter_type](output="sos", N=N, rp=rp, Wn=Wn, btype=btype)

    # Filter the data
    filtered = sosfilt(filter_output, data, axis=0)

    # Return series if that was the input type
    return pd.Series(filtered, index=data.index)



# Copyright 2021 Cognite AS
from .alma import alma
from .arma_smoother import arma
from .butterworth import butterworth
from .chebyshev import chebyshev
from .eweight_ma import ewma
from .lweight_ma import lwma
from .savitzky_golay import sg
from .simple_ma import sma


TOOLBOX_NAME = "Smooth"

__all__ = [
    "sg",
    "alma",
    "arma",
    "butterworth",
    "chebyshev",
    "ewma",
    "lwma",
    "sma",
]

# TODO: groupd smoothers into types: window and frequency



import functools

from typing import List

from indsl import versioning
from indsl.data_quality import extreme
from indsl.detect import drift, ssid, vma
from indsl.filter import status_flag_filter, wavelet_filter
from indsl.not_listed_operations import no_op
from indsl.oil_and_gas import calculate_shutin_interval, productivity_index
from indsl.regression import poly_regression
from indsl.resample import interpolate, resample, resample_to_granularity
from indsl.smooth import alma, arma, butterworth, chebyshev, ewma, lwma, sg, sma
from indsl.statistics import remove_outliers
from indsl.ts_utils import (
    absolute,
    add,
    arccos,
    arccosh,
    arcsin,
    arcsinh,
    arctan,
    arctan2,
    arctanh,
    bin_map,
    ceil,
    clip,
    cos,
    cosh,
    deg2rad,
    differentiate,
    div,
    exp,
    floor,
    inv,
    log,
    log2,
    log10,
    logn,
    maximum,
    minimum,
    mod,
    mul,
    neg,
    power,
    rad2deg,
    round,
    sign,
    sin,
    sinh,
    sqrt,
    sub,
    tan,
    tanh,
    trapezoidal_integration,
)


old_to_new = {
    "SG_SMOOTHER": sg,
    "ALMA_SMOOTHER": alma,
    "ARMA_SMOOTHER": arma,
    "BTR_SMOOTHER": butterworth,
    "CHB_SMOOTHER": chebyshev,
    "EXP_WMA": ewma,
    "LINEAR_WMA": lwma,
    "SIMPLE_MA": sma,
    "INTERPOLATE": interpolate,
    "RESAMPLE_EXTENDED": resample,
    "RESAMPLE": resample_to_granularity,
    "POLY_REGRESSOR": poly_regression,
    "SHUTIN_CALC": calculate_shutin_interval,
    "PI_CALC": productivity_index,
    "PASSTHROUGH": no_op,
    "OUTLIER_DETECTOR": extreme,
    "DRIFT_DETECTOR": drift,
    "SS_DETECTOR": ssid,
    "VARIABLE_MA": vma,
    "WAVELET_FILTER": wavelet_filter,
    "STATUS_FLAG_FILTER": status_flag_filter,
    "OUTLIERS_REMOVE": remove_outliers,
    "ROUND": round,
    "FLOOR": floor,
    "CEIL": ceil,
    "SIGN": sign,
    "CLIP": clip,
    "MAX": maximum,
    "MIN": minimum,
    "BIN_MAP": bin_map,
    "SIN": sin,
    "COS": cos,
    "TAN": tan,
    "ARCSIN": arcsin,
    "ARCCOS": arccos,
    "ARCTAN": arctan,
    "ARCTAN2": arctan2,
    "DEG2RAD": deg2rad,
    "RAD2DEG": rad2deg,
    "SINH": sinh,
    "COSH": cosh,
    "TANH": tanh,
    "ARCSINH": arcsinh,
    "ARCCOSH": arccosh,
    "ARCTANH": arctanh,
    "ADD": add,
    "SUB": sub,
    "MUL": mul,
    "DIV": div,
    "POW": power,
    "INV": inv,
    "SQRT": sqrt,
    "NEG": neg,
    "ABS": absolute,
    "MOD": mod,
    "INTEGRATE": trapezoidal_integration,
    "DDX": differentiate,
    "EXP": exp,
    "LOG": log,
    "LOG2": log2,
    "LOG10": log10,
    "LOGN": logn,
}

for old_op_code, f in old_to_new.items():

    # If multiple versions exist, use version 1.0
    try:
        f = versioning.get(versioning.get_name(f), "1.0")
    except ValueError:
        pass

    # inDSLs versioning does not support giving a function two names
    # (here e.g. SIN and sin). We work around this by creating a
    # dummy-wrapper function
    def make_cpy(f):  # noqa
        return functools.wraps(f)(lambda *args, **kwargs: f(*args, **kwargs))

    f_cpy = make_cpy(f)

    # Register function as a deprecated function with the old op code
    versioning.register(
        "1.0", old_op_code, deprecated=True, changelog="Deprecated function. Re-create node to upgrade"
    )(f_cpy)

    # Attach function to this module
    vars()[old_op_code] = f_cpy

# Expose only old style functions as module members
__all__: List[str] = list(old_to_new.keys())



# Copyright 2021 Cognite AS
from __future__ import annotations

from typing import List, Literal, Optional, Tuple, Union

import pandas as pd

from indsl.data_quality.gaps_identification import (
    gaps_identification_iqr,
    gaps_identification_modified_z_scores,
    gaps_identification_threshold,
    gaps_identification_z_scores,
)
from indsl.data_quality.score.base import DataQualityScore, DataQualityScoreAnalyser
from indsl.type_check import check_types


class GapEventsList(list):
    """List of gap events with additional properties.

    Attributes:
        events: List of gap events.
    """

    def __init__(self, events: List[Tuple[pd.Timestamp, pd.Timestamp]]):
        """Initialize list of gap events."""
        super().__init__(events)

    @property
    def gaps_lengths(self):
        """Duration of gap events."""
        return [event[1] - event[0] for event in self]

    @property
    def _gaps_in_seconds(self):
        return [gap_length.total_seconds() for gap_length in self.gaps_lengths]

    @property
    def gaps_max(self):
        """Maximum gap event duration."""
        return max(self.gaps_lengths) if self.gaps_lengths else None

    @property
    def gaps_avg(self):
        """Average gap event duration."""
        try:
            timedeltas_in_seconds = sum(self._gaps_in_seconds)
            avg_in_seconds = timedeltas_in_seconds / len(self)
        except ZeroDivisionError:
            return None
        return pd.Timedelta(seconds=avg_in_seconds)


class GapDataQualityScoreAnalyser(DataQualityScoreAnalyser):
    """Gap based data quality scores."""

    @check_types
    def __init__(self, series: pd.Series):
        """Gap based data quality scores init function.

        Args:
            series: Series to be analysed

        Raises:
            UserValueError: If the series has less than 2 values
            UserValueError: If series has no time index
        """
        super().__init__(series)
        self._gap_detection_methods = {
            "iqr": gaps_identification_iqr,
            "z_scores": gaps_identification_z_scores,
            "modified_z_scores": gaps_identification_modified_z_scores,
            "threshold": gaps_identification_threshold,
        }

    @check_types
    def get_gaps(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        gap_detection_method: Literal["iqr", "z_scores", "modified_z_scores", "threshold"] = "iqr",
        **gap_detection_options: Optional[Union[pd.Timedelta, int, bool]],
    ) -> GapEventsList:
        """Calculate gap events.

        Args:
            analysis_start: Analyis start time
            analysis_end: Analyis end time
            gap_detection_method: Gap detection method
                Must be one of "iqr", "z_scores", "modified_z_scores"
            gap_detection_options: Arguments to gap detection method
                Provided as a keyword dictionary

        Returns:
            GapEventList: A GapEventList object
        """
        method = self._gap_detection_methods[gap_detection_method]
        gaps = method(self.series, **gap_detection_options)

        events = self._convert_series_to_events(gaps)
        events = self._filter_events_outside_analysis_period(events, analysis_start, analysis_end)
        # The first and last gap might range outside the analysis period. Let's fix this...
        events = self._limit_first_and_last_events_to_analysis_period(events, analysis_start, analysis_end)

        return GapEventsList(events)

    @check_types
    def compute_score(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        gap_detection_method: Literal["iqr", "z_scores", "modified_z_scores", "threshold"] = "iqr",
        **gap_detection_options: Optional[Union[pd.Timedelta, int, bool]],
    ) -> DataQualityScore:
        """Compute the gap analysis score.

        Args:
            analysis_start: Analyis start time
            analysis_end: Analyis end time
            gap_detection_method: Gap detection method
                Must be one of "iqr", "z_scores", "modified_z_scores"
            gap_detection_options: Arguments to gap detection method
                Provided as a keyword dictionary

        Returns:
            DataQualityScore: A GapDataQualityScore object

        Raises:
            UserValueError: If analysis_start < analysis_end
            UserValueError: If the analysis start and end timestamps are outside the range of the series index
        """
        super().compute_score(analysis_start, analysis_end)
        # Treat empty series as one gap
        if len(self.series) == 0:
            return DataQualityScore(analysis_start, analysis_end, [(analysis_start, analysis_end)])

        events = self.get_gaps(analysis_start, analysis_end, gap_detection_method, **gap_detection_options)

        return DataQualityScore(analysis_start, analysis_end, [tuple(event) for event in events])
